# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Plots of Guardian data
"""

from collections import OrderedDict

from six import string_types

from gwpy.segments import (Segment, SegmentList)
from gwpy.plot.colors import tint
from gwpy.plot.segments import SegmentRectangle

from ..data import get_timeseries
from ..segments import get_segments
from ..utils import re_quote
from .registry import (get_plot, register_plot)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


class GuardianStatePlot(get_plot('segments')):
    type = 'guardian'
    defaults = get_plot('segments').defaults.copy()
    defaults.update({
        'color': None,
        'linewidth': 0.5,
        'requestcolor': '#0066ff',
        'nominalcolor': '#ffb200',
        'legend_loc': 'upper left',
        'legend_bbox_to_anchor': (1., 1.),
        'legend_borderaxespad': 0.,
        'legend_frameon': False,
        'legend_fontsize': 12,
        'ytick.labelsize': 10,
    })

    def __init__(self, *args, **kwargs):
        super(GuardianStatePlot, self).__init__(*args, **kwargs)
        self.preview_labels = True

    @property
    def node(self):
        return self.flags[0].split(' ', 1)[0][3:]

    @property
    def ifo(self):
        return list(self.ifos)[0]

    def draw(self):
        from ..tabs.guardian import (REQUESTSTUB, NOMINALSTUB,
                                     re_guardian_index)

        plot = self.init_plot()
        ax = plot.gca()

        # get labels
        flags = [str(f).replace('_', r'\_') for f in self.flags]
        labels = self.pargs.pop('labels', self.pargs.pop('label', flags))
        ax.set_insetlabels(self.pargs.pop('insetlabels', True))
        if isinstance(labels, string_types):
            labels = labels.split(',')
        labels = [re_quote.sub('', str(s).strip('\n ')) for s in labels]

        # parse plotting arguments
        legendargs = self.parse_legend_kwargs()
        plotargs = self.parse_plot_kwargs()[0]
        plotargs.pop('label')
        height = plotargs.pop('height', .8)
        activecolor = plotargs.pop('facecolor')
        nominalcolor = self.pargs.pop('nominalcolor')
        requestcolor = self.pargs.pop('requestcolor')
        actargs = plotargs.copy()
        plotargs.update({
            'facecolor': nominalcolor,
            'edgecolor': 'none',
            'known': {'alpha': 0.1, 'facecolor': 'lightgray'},
        })
        reqargs = plotargs.copy()
        reqargs.update({
            'facecolor': requestcolor,
            'known': None,
            'height': height,
            })
        actargs.update({
            'facecolor': activecolor,
            'known': None,
        })

        if self.state and not self.all_data:
            valid = self.state.active
        else:
            valid = SegmentList([self.span])

        # plot segments
        for y, (flag, label) in enumerate(list(zip(self.flags, labels))[::-1]):
            inreq = str(flag) + REQUESTSTUB
            nominal = str(flag) + NOMINALSTUB
            segs = get_segments([flag, inreq, nominal], validity=valid,
                                query=False)
            # format label
            if self.fileformat != 'svg':
                try:
                    idx, label = re_guardian_index.match(label).groups()
                except AttributeError:
                    pass
                else:
                    x = float(self.span[0]) - (float(abs(self.span)) * 0.005)
                    ax.text(x, y, '[%s]' % idx, ha='right', va='center',
                            fontsize=12)
            # plot segments
            ax.plot(segs[nominal], label=label, y=y, height=1., **plotargs)
            ax.plot(segs[inreq], label=label, y=y, collection='ignore',
                    **reqargs)
            ax.plot(segs[flag], label=label, y=y, collection='ignore',
                    height=.6, **actargs)

        # make custom legend
        seg = Segment(self.start - 10, self.start - 9)
        v = plotargs.pop('known', None)
        n = SegmentRectangle(seg, 0, facecolor=nominalcolor,
                             edgecolor=nominalcolor)
        a = SegmentRectangle(seg, 0, facecolor=requestcolor,
                             edgecolor=requestcolor)
        b = SegmentRectangle(seg, 0, facecolor=activecolor,
                             edgecolor=actargs['edgecolor'])
        handles = [n, a, b]
        labels = ['Nominal', 'Request', 'Active']
        if v:
            v.pop('collection', None)
            v['edgecolor'] = v.get('edgecolor') or tint(v['facecolor'], .5)
            handles.insert(0, SegmentRectangle(seg, 0, **v))
            labels.insert(0, 'Alive')
        ax.legend(handles, labels, title='Node state', **legendargs)

        # customise plot
        for key, val in self.pargs.items():
            try:
                getattr(ax, 'set_%s' % key)(val)
            except AttributeError:
                setattr(ax, key, val)
        if 'ylim' not in self.pargs:
            ax.set_ylim(-0.5, len(self.flags) - 0.5)

        epoch = ax.get_epoch()

        # add node MODE along the bottom
        sax = None
        seg_kw = {'y': 0, 'edgecolor': 'none', 'label': 'Mode'}
        legentry = OrderedDict()
        colors = iter(('#0066ff', '#8000bf', 'hotpink', 'yellow'))
        for ctag, mstate in [
            # bit listing for channels (Nones are ignored)
            ('MODE', ('EXEC', 'MANAGED', 'MANUAL')),
            ('OP', (None, 'PAUSE', None)),
        ]:
            # get data
            data = get_timeseries(
                '{0.ifo}:GRD-{0.node}_{1}'.format(self, ctag), valid,
                query=False).join(gap='pad', pad=-1)
            for i, m in filter(lambda x: x[1] is not None, enumerate(mstate)):
                x = (data == i).to_dqflag()
                fc = next(colors)
                if sax is None:
                    sax = plot.add_segments_bar(
                        x.active, facecolor=fc, **seg_kw)
                else:
                    sax.plot_segmentlist(
                        x.active, facecolor=fc, **seg_kw)
                legentry[m.title()] = SegmentRectangle(seg, 0, facecolor=fc,
                                                       edgecolor=fc)
                seg_kw.update({'collection': 'ignore', 'label': None})

        # add OK segments along the bottom
        ok = get_segments(flag.split(' ', 1)[0] + ' OK', validity=valid,
                          query=False)
        seg_kw.pop('edgecolor', None)
        sax.plot_segmentlist(ok.active, facecolor=activecolor,
                             edgecolor=actargs['edgecolor'],
                             **seg_kw)
        legentry['`OK\''] = SegmentRectangle(seg, 0, facecolor=activecolor,
                                             edgecolor=actargs['edgecolor'])

        # make custom legend
        legendargs.update({
            'bbox_to_anchor': (1., 0.),
            'loc': 'lower left',
        })
        sax.legend(list(legentry.values()), list(legentry),
                   title='Node mode', **legendargs)
        sax.tick_params(axis='y', which='major', labelsize=12)
        for ax_ in (ax, sax):
            ax_.set_epoch(epoch)

        return self.finalize()


register_plot(GuardianStatePlot)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013-2016)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the `Tab` API, and all of the built-in tab objects
"""

# core
from .registry import *
from .core import *
from .builtin import *
from .misc import *

# data
from .data import *

# application-specific extras
from .sei import *
from .guardian import *
from .stamp import *
from .management import *
from .etg import *
from .fscan import *
from .gracedb import *

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""A `Plot` is a representation of an image to be included in the HTML
output a :doc:`tab </tabs>`.

For simple purposes, a `Plot` is just a reference to an existing image file
that can be imported into an HTML page via the ``<img>`` tag.

For more complicated purposes, a number of data plot classes are provided to
allow users to generate images on-the-fly.
The available classes are:

.. autosummary::
   :toctree: api

   TimeSeriesDataPlot
   SpectrogramDataPlot
   SegmentDataPlot
   StateVectorDataPlot
   SpectrumDataPlot
   TimeSeriesHistogramPlot
   TriggerTimeSeriesDataPlot
   TriggerHistogramPlot
   TriggerRateDataPlot
"""

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

from matplotlib import rcParams
from gwpy.plot.tex import (has_tex, MACROS as GWPY_TEX_MACROS)
from gwpy.utils.env import bool_env

from .registry import *
from .utils import *
from .core import *
from .builtin import *
from .segments import *
from .triggers import *
from .range import *
from .noisebudget import *
from .guardian import *
from .sei import *

rcParams.update({
    # custom GWSumm formatting
    'font.size': 10,
    'xtick.labelsize': 18,
    'ytick.labelsize': 18,
    'axes.labelsize': 20,
    'axes.titlesize': 24,
    'grid.alpha': 0.5,
    'figure.figsize': (12, 6),
    'svg.fonttype': 'none',
})

if has_tex() and bool_env("GWPY_USETEX", True):
    rcParams.update({
        # reproduce GWPY_TEX_RCPARAMS
        'text.usetex': True,
        'text.latex.preamble': (
            rcParams.get('text.latex.preamble', []) + GWPY_TEX_MACROS),
        'font.family': ['serif'],
        'axes.formatter.use_mathtext': False,
    })



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for segment handling and display
"""

from __future__ import (division, print_function)
import operator
import sys
import warnings
from collections import OrderedDict
from configparser import (
    DEFAULTSECT,
    ConfigParser,
    NoSectionError,
    NoOptionError,
)

from six import string_types
from six.moves import reduce

from astropy.io.registry import IORegistryError

from gwpy.segments import (DataQualityFlag, DataQualityDict,
                           SegmentList, Segment)

from . import globalv
from .utils import (
    re_flagdiv,
    vprint,
    WARNC,
    ENDC,
)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

SEGDB_URLS = [
    'https://segdb.ligo.caltech.edu',
    'https://metaserver.phy.syr.edu',
    'https://geosegdb.atlas.aei.uni-hannover.de',
    'http://10.20.50.30'  # geosegdb internal
]


def get_segments(flag, validity=None, config=ConfigParser(), cache=None,
                 query=True, return_=True, coalesce=True, padding=None,
                 segdb_error='raise', url=None):
    """Retrieve the segments for a given flag

    Segments will be loaded from global memory if already defined,
    otherwise they will be loaded from the given
    :class:`~glue.lal.Cache`, or finally from the segment database

    Parameters
    ----------
    flag : `str`, `list`
        either the name of one flag, or a list of names

    validity : `~gwpy.segments.SegmentList`
        the segments over which to search for other segments

    query : `bool`, optional, default: `True`
        actually execute a read/query operation (if needed), otherwise
        just retrieve segments that have already been cached

    config : `~configparser.ConfigParser`, optional
        the configuration for your analysis, if you have one. If
        present the ``[segment-database]`` section will be queried
        for the following options

        - ``gps-start-time``, and ``gps-end-time``, if ``validity`` is
          not given
        - ``url`` (the remote hostname for the segment database) if
          the ``url`` keyword is not given

    cache : :class:`glue.lal.Cache`, optional
        a cache of files from which to read segments, otherwise segments
        will be downloaded from the segment database

    coalesce : `bool`, optional, default: `True`
        coalesce all segmentlists before returning, otherwise just return
        segments as they were downloaded/read

    padding : `tuple`, or `dict` of `tuples`, optional
        `(start, end)` padding with which to pad segments that are
        downloaded/read

    segdb_error : `str`, optional, default: ``'raise'``
        how to handle errors returned from the segment database, one of

        - ``'raise'`` (default) : raise the exception as normal
        - ``'warn'`` : print the exception as a warning, but return no
          segments
        - ``'ignore'`` : silently ignore the error and return no segments

    url : `str`, optional
        the remote hostname for the target segment database

    return_ : `bool`, optional, default: `True`
        internal flag to enable (True) or disable (False) actually returning
        anything. This is useful if you want to download/read segments now
        but not use them until later (e.g. plotting)

    Returns
    -------
    flag : `~gwpy.segments.DataQualityFlag`
        the flag object representing segments for the given single flag, OR

    flagdict : `~gwpy.segments.DataQualityDict`
        the dict of `~gwpy.segments.DataQualityFlag` objects for multiple
        flags, if ``flag`` is given as a `list`, OR

    None
       if ``return_=False``
    """
    if isinstance(flag, string_types):
        flags = flag.split(',')
    else:
        flags = flag
    allflags = set([f for cf in flags for f in
                    re_flagdiv.split(str(cf))[::2] if f])

    if padding is None and isinstance(flag, DataQualityFlag):
        padding = {flag: flag.padding}
    elif padding is None:
        padding = dict((flag,
                        isinstance(flag, DataQualityFlag) and
                        flag.padding or None) for flag in flags)

    # check validity
    if validity is None:
        start = config.get(DEFAULTSECT, 'gps-start-time')
        end = config.get(DEFAULTSECT, 'gps-end-time')
        span = SegmentList([Segment(start, end)])
    elif isinstance(validity, DataQualityFlag):
        validity = validity.active
        try:
            span = SegmentList([validity.extent()])
        except ValueError:
            span = SegmentList()
    else:
        try:
            span = SegmentList([SegmentList(validity).extent()])
        except ValueError:
            span = SegmentList()
    validity = SegmentList(validity)

    # generate output object
    out = DataQualityDict()
    for f in flags:
        out[f] = DataQualityFlag(f, known=validity, active=validity)
    for f in allflags:
        globalv.SEGMENTS.setdefault(f, DataQualityFlag(f))

    # read segments from global memory and get the union of needed times
    try:
        old = reduce(
            operator.and_,
            (globalv.SEGMENTS.get(f, DataQualityFlag(f)).known for f in flags))
    except TypeError:
        old = SegmentList()
    newsegs = validity - old
    # load new segments
    query &= abs(newsegs) != 0
    query &= len(allflags) > 0
    if cache is not None:
        query &= len(cache) != 0
    if query:
        if cache is not None:
            try:
                new = DataQualityDict.read(cache, list(allflags))
            except IORegistryError as e:
                # can remove when astropy >= 1.2 is required
                if type(e) is not IORegistryError:
                    raise
                if len(allflags) == 1:
                    f = list(allflags)[0]
                    new = DataQualityDict()
                    new[f] = DataQualityFlag.read(cache, f, coalesce=False)
            for f in new:
                new[f].known &= newsegs
                new[f].active &= newsegs
                if coalesce:
                    new[f].coalesce()
                vprint("    Read %d segments for %s (%.2f%% coverage).\n"
                       % (len(new[f].active), f,
                          float(abs(new[f].known))/float(abs(newsegs))*100))
        else:
            if len(newsegs) >= 10:
                qsegs = span
            else:
                qsegs = newsegs
            # parse configuration for query
            kwargs = {}
            if url is not None:
                kwargs['url'] = url
            else:
                try:
                    kwargs['url'] = config.get('segment-database', 'url')
                except (NoSectionError, NoOptionError):
                    pass
            if kwargs.get('url', None) in SEGDB_URLS:
                query_func = DataQualityDict.query_segdb
            else:
                query_func = DataQualityDict.query_dqsegdb
            try:
                new = query_func(allflags, qsegs, on_error=segdb_error,
                                 **kwargs)
            except Exception as e:
                # ignore error from SegDB
                if segdb_error in ['ignore', None]:
                    pass
                # convert to warning
                elif segdb_error in ['warn']:
                    print('%sWARNING: %sCaught %s: %s [gwsumm.segments]'
                          % (WARNC, ENDC, type(e).__name__, str(e)),
                          file=sys.stderr)
                    warnings.warn('%s: %s' % (type(e).__name__, str(e)))
                # otherwise raise as normal
                else:
                    raise
                new = DataQualityDict()
            for f in new:
                new[f].known &= newsegs
                new[f].active &= newsegs
                if coalesce:
                    new[f].coalesce()
                vprint("    Downloaded %d segments for %s (%.2f%% coverage).\n"
                       % (len(new[f].active), f,
                          float(abs(new[f].known))/float(abs(newsegs))*100))
        # record new segments
        globalv.SEGMENTS += new
        for f in new:
            globalv.SEGMENTS[f].description = str(new[f].description)

    # return what was asked for
    if return_:
        for compound in flags:
            union, intersection, exclude, notequal = split_compound_flag(
                compound)
            if len(union + intersection) == 1:
                out[compound].description = globalv.SEGMENTS[f].description
                out[compound].padding = padding.get(f, (0, 0))
            for flist, op in zip([exclude, intersection, union, notequal],
                                 [operator.sub, operator.and_, operator.or_,
                                  not_equal]):
                for f in flist:
                    pad = padding.get(f, (0, 0))
                    segs = globalv.SEGMENTS[f].copy()
                    if isinstance(pad, (float, int)):
                        segs = segs.pad(pad, pad)
                    elif pad is not None:
                        segs = segs.pad(*pad)
                    if coalesce:
                        segs = segs.coalesce()
                    out[compound] = op(out[compound], segs)
            out[compound].known &= validity
            out[compound].active &= validity
            if coalesce:
                out[compound].coalesce()
        if isinstance(flag, string_types):
            return out[flag]
        else:
            return out


def not_equal(a, b, f):
    diff1 = a - b
    diff2 = b - a
    return diff1 | diff2


def split_compound_flag(compound):
    """Parse the configuration for this state.

    Returns
    -------
    flags : `tuple`
        a 2-tuple containing lists of flags defining required ON
        and OFF segments respectively for this state
    """
    # find flags
    divs = re_flagdiv.findall(compound)
    keys = re_flagdiv.split(compound)
    # load flags and vetoes
    union = []
    intersection = []
    exclude = []
    notequal = []
    for i, key in enumerate(keys[::2]):
        if not key:
            continue
        # get veto bit
        if i != 0 and divs[i-1] == '!':
            exclude.append(key)
        elif i != 0 and divs[i-1] == '|':
            union.append(key)
        elif i != 0 and divs[i-1] == '!=':
            notequal.append(key)
        else:
            intersection.append(key)
    return union, intersection, exclude, notequal


def format_padding(flags, padding):
    """Format an arbitrary collection of paddings into a `dict`
    """
    # parse string to start with
    if isinstance(padding, string_types):
        padding = list(eval(str))
    # zip list into dict
    if (isinstance(padding, (list)) or
            (isinstance(padding, tuple) and len(padding) and
             (any(isinstance(p, (list, tuple)) for p in padding) or
              len(padding) > 2))):
        return OrderedDict(list(zip(flags, padding)))
    # otherwise copy single padding param for all flags
    elif not isinstance(padding, dict):
        return OrderedDict((c, padding) for c in flags)
    else:
        return padding



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.


"""Parse, define, and geenrate plots as requests through the configuration
for GWSumm
"""

import os.path
import re
import warnings
from collections import OrderedDict
from math import (floor, ceil)
from numbers import Number

from six import string_types
from six.moves.urllib.parse import urlparse

from matplotlib import (rcParams, rc_context)

from gwpy.segments import Segment
from gwpy.detector import ChannelList
from gwpy.plot import Plot

from ..channels import (get_channel, split as split_channels,
                        split_combination as split_channel_combination)
from ..config import GWSummConfigParser
from ..state import get_state
from ..utils import (vprint, safe_eval, re_quote)
from . import utils as putils
from .registry import register_plot

__all__ = ['SummaryPlot', 'DataPlot']

re_cchar = re.compile(r"[\W\s_]+")

putils.AXES_PARAMS.extend([
    'insetlabels',  # for segment plotting
])
NON_PLOT_PARAMS = set(putils.FIGURE_PARAMS + putils.AXES_PARAMS)


# -- utilities ----------------------------------------------------------------

def format_label(label):
    label = str(label).strip('\n ')
    label = re_quote.sub('', label)
    return putils.usetex_tex(label)


# -- basic Plot object --------------------------------------------------------

class SummaryPlot(object):
    """An image to displayed in GWSumm HTML output.

    Parameters
    ----------
    href : `str`, optional
        The IMG URL for this `SummaryPlot`.
    new : `bool`, optional
        `bool` flag whether this is a new plot to be processed (`True`),
        of that the output already exists on disk (`False`).

    Notes
    -----
    This `class` is a stub, designed to make creating detailed `SummaryPlot`
    classes easier.
    """
    type = None
    _threadsafe = True

    def __init__(self, href=None, src=None, new=True, caption=''):
        self.href = href
        if src:
            self.src = src
        self.new = new
        self.caption = caption

    @property
    def href(self):
        """HTML <img> href attribute for this `SummaryPlot`.
        """
        return self._href

    @href.setter
    def href(self, url):
        if url is None:
            self._href = None
        elif urlparse(url).netloc:
            self._href = url
        else:
            self._href = os.path.normpath(url)

    @property
    def new(self):
        """Flag whether this is a new plot or, already exists.

        Set new=False to skip actually processing this `SummaryPlot`, and
        just link to the outputfile.
        """
        return self._new

    @new.setter
    def new(self, isnew):
        self._new = bool(isnew)

    @property
    def src(self):
        try:
            return self._src
        except AttributeError:
            return self.href

    @src.setter
    def src(self, url):
        self._src = url

    @property
    def caption(self):
        """HTML <fancybox plot> title attribute for this `SummaryPlot`.
        """
        return self._caption

    @caption.setter
    def caption(self, text):
        self._caption = text

    # -- SummaryPlot methods -----------------------

    @classmethod
    def from_ini(cls, *args, **kwargs):
        """Define a new `SummaryPlot` from a an INI-format `ConfigParser`
        section.
        """
        raise NotImplementedError("Sub-classes should provide this method.")

    def __eq__(self, other):
        """Compare this `SummaryPlot` to another.

        Returns
        -------
        `True`
            if the metadata for ``self`` match those of ``other``
        `False`
            otherwise
        """
        if not isinstance(other, self.__class__):
            return False
        if not self.href == other.href:
            return False
        return True

    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, self.href)

    def __str__(self):
        return str(self.href)


register_plot(SummaryPlot)


class DataPlot(SummaryPlot):
    """A `SummaryPlot` from instrumental data.

    Parameters
    ----------
    channels : `list`
        a list of channel names that define the data sources for this
        `DataPlot`
    start : `float`
        GPS start time of this `DataPlot`.
    end : `float`
        GPS end time of this `DataPlot`.
    tag : `str`
        a descriptive tag for this `DataPlot`, used as part of the output
        file name
    outdir : `str`
        output directory path for this `DataPlot`, defaults to
        the current directory
    href : `str`
        custom URL for this plot to link towards.
    **kwargs
        all other keyword arguments to be passed to this plot's
        :meth:`process` method.

    Notes
    -----
    All sub-classes of this object must provide the following methods

    =======================  ==============================================
    :meth:`add_data_source`  routine for appending data sources to the plot
    =======================  ==============================================
    """
    #: name for DataPlot subclass
    type = 'data'
    #: plot() call style
    _single_call = False
    #: dict of default plotting kwargs
    defaults = {}
    #: list of parameters parsed for `plot()` calls
    DRAW_PARAMS = list(putils.ARTIST_PARAMS)

    def __init__(self, channels, start, end, state=None, outdir='.',
                 tag=None, pid=None, href=None, new=True, all_data=False,
                 read=True, fileformat='png', caption=None, **pargs):
        super(DataPlot, self).__init__(href=href, new=new, caption=caption)
        if isinstance(channels, string_types):
            channels = split_channels(channels)
        self.channels = channels
        self.span = (start, end)
        self.state = state
        self._outdir = outdir
        if tag is not None:
            self.tag = tag
        if pid is not None:
            self.pid = pid
        # allow user to specify all-data instead of all_data as kwarg
        # mainly for INI-parsing convenience, should fix properly
        self.all_data = pargs.pop('all-data', all_data)
        self.pargs = self.defaults.copy()
        self.pargs.update(pargs)
        self.parse_rcParams(self.pargs)
        self.plot = None
        self.read = read
        self.fileformat = fileformat

    # -- properties -----------------------------

    @property
    def span(self):
        """The GPS [start, stop) interval for this `DataPlot`.
        """
        return self._span

    @span.setter
    def span(self, seg):
        self._span = Segment(*seg)

    @property
    def start(self):
        return self.span[0]

    @property
    def end(self):
        return self.span[1]

    @property
    def state(self):
        """`~gwsumm.state.SummaryState` defining validity of this
        `DataPlot`.
        """
        return self._state

    @state.setter
    def state(self, state_):
        if isinstance(state_, string_types):
            self._state = get_state(state_)
        else:
            self._state = state_

    @property
    def channels(self):
        """List of data-source
        :class:`Channels <~gwpy.detector.channel.Channel>` for this
        `DataPlot`.

        :type: :class:`~gwpy.detector.channel.ChannelList`
        """
        return ChannelList(get_channel(c) for c in self._channels)

    @channels.setter
    def channels(self, channellist):
        self._channels = channellist

    @property
    def allchannels(self):
        """List of all unique channels for this plot
        """
        return type(self.channels)(OrderedDict.fromkeys(
            c2 for c in self.channels for
            c2 in split_channel_combination(c.ndsname)
        ).keys())

    @property
    def ifos(self):
        """Interferometer set for this `DataPlot`
        """
        return set([c.ifo for c in self.allchannels if c.ifo])

    @property
    def tag(self):
        """File tag for this `DataPlot`.
        """
        try:
            return self._tag
        except AttributeError:
            state = re_cchar.sub(
                '_',
                self.state is None and 'MULTI' or self.state.name).rstrip('_')
            type_ = re_cchar.sub('_', self.type)
            self._tag = '_'.join([state, self.pid, type_]).upper()
            return self.tag

    @tag.setter
    def tag(self, filetag):
        if filetag is None:
            del self.tag
        else:
            self._tag = filetag

    @tag.deleter
    def tag(self):
        del self._tag

    @property
    def pid(self):
        try:
            return self._pid
        except AttributeError:
            chans = "".join(map(str, self.channels))
            filts = "".join(map(str, [
                getattr(c, 'filter', getattr(c, 'frequency_response', ''))
                for c in self.channels]))
            self._pid = putils.hash(chans + filts)
            return self.pid

    @pid.setter
    def pid(self, id_):
        self._pid = str(id_)

    @pid.deleter
    def pid(self):
        del self._pid

    @property
    def outputfile(self):
        """Output file for this `DataPlot`.
        """
        ifos = ''.join(sorted(self.ifos))
        tag = self.tag
        gps = floor(self.start)
        dur = ceil(self.end - self.start)
        return os.path.join(
            self._outdir,
            '%s-%s-%d-%d.%s' % (ifos, tag, gps, dur, self.fileformat))

    @property
    def href(self):
        if self._href is None:
            return self.outputfile
        else:
            return self._href

    @href.setter
    def href(self, url):
        self._href = url and os.path.normpath(url) or None

    # -- read-only plot properties --------------

    def _is_log(self, axis):
        scale = '{0}scale'.format(axis)
        try:
            return self.pargs[scale] == 'log'
        except KeyError:
            try:
                return self.pargs['log{0}'.format(axis)]
            except KeyError:
                try:
                    ax = self.plot.gca()
                except AttributeError:  # plot not generated yet
                    return False
                return getattr(ax, 'get_{0}'.format(scale))() == 'log'

    @property
    def logx(self):
        return self._is_log('x')

    @property
    def logy(self):
        return self._is_log('y')

    # -- basic methods --------------------------

    def add_channel(self, channel):
        self._channels.append(channel)

    def get_channel_groups(self):
        """Find and group (mean, min, max) sets of channels for plotting.

        Returns
        -------
        groups : `list` of `tuple`
            list of (channelname, channellist) tuples giving core channel
            name and an ordered list of channels. Ordering in preference
            of 'rms', 'mean', 'min', 'max'.

        Notes
        -----
        This method used to return an `OrderedDict`, but was changed to
        return a `list` of `tuple` to enable plotting a channel multiple
        times on a plot, for whatever reason.
        """
        all_ = self.channels
        out = []
        for c in all_:
            if c.ifo == 'G1' and re.search(r'-(av|min|max)\Z', c.texname):
                name = c.texname.rsplit('-', 1)[0]
            else:
                name = c.texname.rsplit('.', 1)[0]
            if ' ' in c.texname:
                out.append((c.texname, [c]))
            else:
                try:
                    id_ = list(zip(*out))[0].index(name)
                except (IndexError, ValueError):
                    out.append((name, [c]))
                else:
                    out[id_][1].append(c)
        order = ['rms', 'mean', 'av', 'min', 'max']
        for channel, clist in out:
            clist.sort(key=lambda c: c.name.split('.')[-1] in order and
                       order.index(c.name.split('.')[-1])+1 or 10)
        return out

    @classmethod
    def from_ini(cls, config, section, start, end, channels=None, **kwargs):
        """Define a new `DataPlot`.
        """
        config = GWSummConfigParser.from_configparser(config)

        # read parameters
        try:
            params = dict(config.nditems(section))
        except AttributeError:
            params = dict(config.items(section))

        # get and check type
        ptype = re.sub(r'[\'\"]', '', params.pop('type'))
        if ptype != cls.type:
            warnings.warn("'%s' plot definition from configuration being "
                          "parsed by different plotting class '%s'"
                          % (ptype, cls.__name__))
        # get channels
        if channels is None:
            channels = params.pop('channels', [])
        if isinstance(channels, string_types):
            channels = split_channels(channels)
        # parse specific parameters
        if 'all-data' in params:
            params['all_data'] = params.pop('all-data')

        # parse other parameters
        for key, val in params.items():
            params[key] = safe_eval(val)
        params.update(kwargs)

        # format and return
        return cls(channels, start, end, **params)

    # -- plot parameter parsing -----------------

    def _parse_param(self, pdict, key, allow_plural=False):
        """Parse a configuration parameter for this Plot from a dict

        Parameters
        ----------
        pdict : `dict`
            the dict to evaluate from

        key : `str`
            the key to evaluated

        allow_plural : `bool`, optional
            try to find plural version of ``key`` if singular not found,
            i.e. ``'colors'`` instead of ``'color'``, default is `False`

        Returns
        -------
        values : `list`
            a mapping of the parsed value for each channel in this plot

        Raises
        ------
        KeyError
            if ``key`` is not found in ``pdict``

        Notes
        -----
        This method uses `gwsumm.utils.safe_eval` to `eval` strings into
        python objects.

        Examples
        --------
        Consider a `DataPlot` with two channels to display

        >>> from gwsumm.plot import DataPlot
        >>> a = DataPlot(['channel1', 'channel2'], 0, 1)

        We can then parse params as follows:

        >>> a._parse_param({'marker': 'x'}, 'marker')
        ['x', 'x']
        >>> a._parse_param({'linestyles': ['-', '--']}, 'linestyle'}
        ['-', '--']
        >>> a._parse_param({'colors': "'red','green'"}, 'color'}
        ['red', 'green']
        """
        # parse keyword
        try:
            val = pdict.pop(key)
        except KeyError as e:
            if not allow_plural:
                raise
            # check for plural
            try:
                val = pdict.pop('%ss' % key)
            except KeyError:
                raise e

        # evaluate (safely) allowing references to self as 'plot'
        if isinstance(val, string_types):
            try:
                val = safe_eval(val, locals_={'plot': self, 'self': self})
            except ZeroDivisionError:  # e.g. zero livetime
                val = 0

        # don't use sets
        if isinstance(val, set):
            val = list(val)

        # if a single-call style plot, just return the value as given
        if self._single_call:
            return val

        # otherwise convert to a 1<->1 mapping with the channels list
        nchans = len(self.get_channel_groups())
        if not isinstance(val, (list, tuple)) or len(val) != nchans:
            return [val] * nchans

        return val

    def _parse_extra_params(self, prefix, **defaults):
        """Parse parameters for an extra plot element

        Parameters
        ----------
        prefix : `str`
            the text prefix identifying parameters for the extra element

        **defaults
            any default options to use

        Returns
        -------
        params : `dict`
        """
        re_prefix = re.compile(r'\A%s[-_]' % prefix.rstrip('-_'))
        extras = defaults.copy()
        for key in list(self.pargs):
            m = re_prefix.match(key)
            if m:
                extras[key[m.span()[1]:]] = safe_eval(self.pargs.pop(key))
        return extras

    def parse_legend_kwargs(self, **defaults):
        """Pop the legend arguments from the `pargs` for this `Plot`
        """
        return self._parse_extra_params('legend', **defaults)

    def parse_plot_kwargs(self, **defaults):
        """Pop keyword arguments for `Axes.plot` from the `pargs` for this Plot
        """
        plotargs = defaults.copy()
        plotargs.setdefault('label', self._parse_labels())

        # loop over known Axes.plot kwargs and parse
        for kwarg in self.DRAW_PARAMS:
            try:
                plotargs[kwarg] = self._parse_param(
                    self.pargs, kwarg, allow_plural=True)
            except KeyError:
                pass

        # normalise log scale parameters
        # TODO: this can be removed once existing config files have been
        #       updated to not use logx and logy options
        for axis in ('x', 'y'):
            logp = 'log{}'.format(axis)
            try:
                log = self.pargs.pop(logp)
            except KeyError:
                continue
            if log is not self._is_log(axis):
                scale = 'log' if log else 'linear'
                self.pargs['{}scale'.format(axis)] = scale

        # if this plot is a single-call plot (where all objects get plotted
        # in a single call out to a ax.plot()-style method) just return
        # the params as a dict of lists
        if self._single_call:
            return plotargs

        # otherwise, map to a list of dicts (one per channel)
        out = []
        nchans = len(self.get_channel_groups())
        for i in range(nchans):
            out.append(dict((key, val[i]) for key, val in plotargs.items() if
                            val is not None and val[i] is not None))
        return out

    def _parse_labels(self, defaults=None):
        """Pop the labels for plotting from the `pargs` for this Plot
        """
        # set default label to show channel name
        chans = list(zip(*self.get_channel_groups()))[0]
        if defaults is None:
            defaults = chans

        # parse user labels
        try:
            labels = self._parse_param(self.pargs, 'label', allow_plural=True)
        except KeyError:
            labels = defaults

        if list(set(labels)) == [labels[0]] and labels[0] is not None:
            labels = labels[0].split(',')

        # escape underscores
        labels = list(map(format_label, labels))

        # fill gaps with None
        while len(labels) < len(chans):
            labels.append(None)

        return labels

    def parse_rcParams(self, params):
        """Parse matplotlib rcParams settings from a dict of plot params
        """
        self.rcParams = {}
        for key in list(params):
            if key in rcParams:
                self.rcParams[key] = safe_eval(params.pop(key))
        return self.rcParams

    def parse_list(self, prefix, **defaults):
        """Parse a list of something from parameters

        This enables listing `hline`s (for example) in the config as

        [plot-blah]
        hline = 100
        hline-linestyle = '--'
        hline-color = 'red'
        hline2 = 200
        hline2-linestyle = '--'
        hline2-color = 'blue'

        Returns an `OrderedDict` with keys matching the primary parsed
        value, and values as everything else, e.g.

        {100: {'linestyle': '--', 'color': 'red'},
         200: {'linestyle': '--', 'color': 'blue'},}
        """
        items = OrderedDict()
        re_prefix = re.compile(r'{0}(\d+)?\Z'.format(prefix))
        keys = sorted(list(self.pargs))
        while True:
            for i, key in enumerate(keys):
                if re_prefix.match(key):
                    primary = safe_eval(self.pargs.pop(key))
                    items[primary] = self._parse_extra_params(key, **defaults)
                    break  # go to next iteration
            else:  # no references matched, so stop
                break
            keys = sorted(list(self.pargs)[i:])

        return items

    # -- figure processing ----------------------

    def process(self, outputfile=None, close=True):
        with rc_context(rc=self.rcParams):
            return self.draw()

    def draw(self):
        """Process all data and generate the output file for this
        `SummaryPlot`.

        This function should be provided by all sub-classes, and should
        take no arguments.
        """
        raise NotImplementedError("This method should be provided by a "
                                  "sub-class")

    def init_plot(self, data=[], FigureClass=Plot, geometry=(1, 1),
                  projection='rectilinear', sharex=True, sharey=True,
                  **kwargs):
        """Initialise the Figure and Axes objects for this `DataPlot`.
        """
        # update plot defaults using channel data
        self._update_defaults_from_channels()

        # strip figure and axes params from pargs
        for key in NON_PLOT_PARAMS:
            try:
                kwargs.setdefault(key, self.pargs.pop(key))
            except KeyError:
                continue
            # escape text for TeX
            if key in ('title', 'xlabel', 'ylabel'):
                kwargs[key] = putils.usetex_tex(kwargs[key])

        # create figure
        self.plot = FigureClass(*data, geometry=geometry,
                                projection=projection, sharex=sharex,
                                sharey=sharey, **kwargs)
        return self.plot

    def _update_defaults_from_channels(self):
        """Update default plotting params from channel attributes

        This method is called at the start of DataPlot.init_plot(),
        so should be populated in any subclasses that want it
        """
        pass

    def finalize(self, outputfile=None, close=True, **savekwargs):
        """Save the plot to disk and close.
        """
        # save figure and close (build both png and pdf for pdf choice)
        if outputfile is None:
            outputfile = self.outputfile
        if not isinstance(outputfile, string_types):
            extensions = [None]
        elif outputfile.endswith('.pdf'):
            extensions = ['.pdf', '.png']
        else:
            extensions = [os.path.splitext(outputfile)[1]]

        for ext in extensions:
            try:
                fp = '%s%s' % (os.path.splitext(outputfile)[0], ext)
            except AttributeError:
                fp = outputfile
            try:
                self.plot.save(fp, **savekwargs)
            except (IOError, RuntimeError, IndexError) as e:
                warnings.warn("Caught %s: %s [retrying...]"
                              % (type(e).__name__, str(e)))
                self.plot.save(fp, **savekwargs)
            if isinstance(fp, string_types):
                vprint("        %s written\n" % fp)
        if close:
            self.plot.close()
        return outputfile

    def apply_parameters(self, *axes, **pargs):
        keys = sorted(list(pargs),
                      key=lambda x: 1 if x in ('xscale', 'yscale') else 2)
        for ax in axes:
            for key in keys:
                if key.startswith('no-'):  # skip no-xxx keys
                    continue
                val = pargs[key]
                if key in ['xlim', 'ylim'] and isinstance(val, string_types):
                    val = eval(val)
                elif key == 'grid':
                    self._apply_grid_params(ax, val)
                    continue
                try:
                    getattr(ax, 'set_%s' % key)(val)
                except AttributeError:
                    setattr(ax, key, val)

    def _apply_grid_params(self, ax, val):
        if val in ('major', 'minor'):
            ax.grid(True, which=val)
        else:
            ax.grid(val)

    def add_hvlines(self):
        """Add horizontal and vertical lines to this `DataPlot`

        These should be defined in the configuration via the `hline` and
        `vline` keys.
        """
        for key in ('hline', 'vline'):
            lines = self.parse_list(key, linestyle='--', color='red')
            for ax in self.plot.axes:
                axline = getattr(ax, 'ax{0}'.format(key))
                for val, params in lines.items():
                    if isinstance(val, Number):
                        val = [val]
                    for x in val:
                        axline(x, **params)


register_plot(DataPlot)


# -- custom plot types --------------------------------------------------------

class BarPlot(DataPlot):
    """`DataPlot` with bars
    """
    type = 'bar'
    DRAW_PARAMS = ['width', 'bottom', 'color', 'edgecolor', 'linewidth',
                   'xerr', 'yerr', 'ecolor', 'capsize', 'error_kw',
                   'align', 'orientation', 'log', 'alpha', 'rasterized']


class PiePlot(DataPlot):
    type = 'pie'
    DRAW_PARAMS = ['explode', 'colors', 'autopct', 'pctdistance', 'shadow',
                   'labeldistance', 'startangle', 'radius', 'counterclock',
                   'wedgeprops', 'textprops', 'rasterized']



# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""`SummaryTab` for seismic watchdog monitoring
"""

import re
from configparser import NoOptionError

from matplotlib.pyplot import subplots
from matplotlib.ticker import NullLocator

from gwpy.plot import Plot
from gwpy.timeseries import TimeSeriesDict

from ..channels import get_channel
from ..utils import re_quote
from .registry import (get_plot, register_plot)


class SeiWatchDogPlot(get_plot('data')):
    """Plot a specific SEI WatchDog trip
    """
    type = 'watchdog'
    data = 'watchdog'

    def __init__(self, gpstime, chamber, sensor, config, outfile, ifo=None,
                 duration=30, nds=False, datacache=None):
        """Configure a new `SeiWatchDogPlot`.
        """
        super(SeiWatchDogPlot, self).__init__([],
                                              int(gpstime) - duration/2.,
                                              int(gpstime) + duration/2.)

        # get params
        if ifo is None:
            ifo = config.get('DEFAULT', 'IFO')
        self.ifo = ifo
        self.chamber = chamber
        self.sensor = sensor
        self.gpstime = gpstime
        self.duration = duration
        self.outputfile = outfile
        self.use_nds = nds

        system = (sensor.split(' ')[0] == 'HEPI' and
                  'HPI' or sensor.split(' ')[0])

        # get channels
        mapsec = 'sei-wd-map-%s' % sensor
        if not config.has_section(mapsec) and re.match(r'ISI ST\d ', sensor):
            mapsec = ('sei-wd-map-%s'
                      % (' '.join(sensor.split(' ', 2)[::2])))
        stubs = list(zip(*sorted(
            [o for o in config.items(mapsec) if o[0].isdigit()],
            key=lambda x: x[0],
        )))[1]
        if re.search(r'ISI ST\d ', sensor):
            stage = sensor.split(' ')[1]
            channels = [get_channel('%s:%s-%s_%s_%s'
                                    % (ifo, system, chamber, stage, stub))
                        for stub in stubs]
        else:
            channels = [get_channel('%s:%s-%s_%s'
                                    % (ifo, system, chamber, stub))
                        for stub in stubs]

        # set types
        for channel in channels:
            if not hasattr(channel, 'type') or not channel.type:
                channel.ctype = 'adc'

        self.chanlist = channels

        try:
            self.geometry = list(map(int, config.get(mapsec, 'geometry').split(',')))
        except NoOptionError:
            self.geometry = (len(channels), 1)
        if len(self.chanlist) != self.geometry[0] * self.geometry[1]:
            raise ValueError("Geometry does not match number of channels.")

        try:
            self.unit = '[%s]' % re_quote.sub('', config.get(mapsec, 'unit'))
        except NoOptionError:
            self.unit = ''

    @property
    def outputfile(self):
        return self._outputfile

    @outputfile.setter
    def outputfile(self, outfile):
        self._outputfile = outfile

    def draw(self):

        # data span
        start = self.gpstime - self.duration / 2.
        end = self.gpstime + self.duration / 2.

        # get data
        if self.use_nds:
            data = TimeSeriesDict.fetch(self.chanlist, start, end)
        else:
            from glue.datafind import GWDataFindHTTPConnection
            conn = GWDataFindHTTPConnection()
            cache = conn.find_frame_urls(self.ifo[0], '%s_R' % self.ifo,
                                         self.start, self.end, urltype='file')
            if len(cache) == 0:
                data = {}
            else:
                data = TimeSeriesDict.read(cache, self.chanlist, start=start,
                                           end=end)

        # make plot
        plot, axes = subplots(nrows=self.geometry[0], ncols=self.geometry[1],
                              sharex=True, subplot_kw={'xscale': 'auto-gps'},
                              FigureClass=Plot, figsize=[12, 6])
        axes[0, 0].set_xlim(start, end)
        for channel, ax in zip(self.chanlist, axes.flat):
            ax.set_epoch(self.gpstime)
            # plot data
            try:
                ax.plot(data[channel])
            except KeyError:
                ax.text(self.gpstime, 0.5, "No data", va='center', ha='center',
                        transform=ax.transData)
            # plot trip indicator
            ax.axvline(self.gpstime, linewidth=0.5, linestyle='--',
                       color='red')
            ax.set_xlabel('')
            ax.set_ylabel('')
            ax.set_title(channel.texname, fontsize=10)
            ax.xaxis.set_minor_locator(NullLocator())
            for tick in ax.yaxis.get_major_ticks():
                tick.label.set_fontsize(10)
            for tick in ax.xaxis.get_major_ticks():
                tick.label.set_fontsize(16)
        plot.text(0.5, 0.02, 'Time [seconds] from trip (%s)' % self.gpstime,
                  ha='center', va='bottom', fontsize=24)
        plot.text(0.01, 0.5, 'Amplitude %s' % self.unit, ha='left',
                  va='center', rotation='vertical', fontsize=24)

        plot.suptitle('%s %s %s watchdog trip: %s'
                      % (self.ifo, self.chamber, self.sensor, self.gpstime),
                      fontsize=24)

        plot.save(self.outputfile)
        plot.close()
        return self.outputfile


register_plot(SeiWatchDogPlot)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013-2016)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Methods and classes for loading and pre-processing data

Each of the sub-modules are designed to read or create the data requested
only once, with the containers from the `globalv` module used as a storage
buffer for each unique data type
"""

# read TimeSeries data
from .timeseries import *

# generate Spectrograms and FrequencySeries
from .spectral import *

# generate coherence Spectrograms and Spectra
from .coherence import *

# generate TimeSeries of sensitive distance (range)
from .range import *



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""HTML <head> helphers

This module mainly declares the resources used by standard on HTML pages
"""

import os.path
from collections import OrderedDict
from gwdetchar.io.html import (CSS_FILES, JS_FILES)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


STATICDIR = os.path.join(os.path.dirname(__file__), 'static')

# build list of javascript resources
JS = OrderedDict((
    ('jquery', JS_FILES[0]),
    ('moment', JS_FILES[1]),
    ('bootstrap', JS_FILES[2]),
    ('fancybox', JS_FILES[3]),
    ('datepicker', (
        '//cdnjs.cloudflare.com/ajax/libs/bootstrap-datepicker/'
        '1.6.0/js/bootstrap-datepicker.min.js')),
    ('bootstrap-ligo', JS_FILES[4]),
    ('gwdetchar', JS_FILES[5]),
    ('gwsumm', os.path.join(STATICDIR, 'gwsumm.min.js')),
))

# build list of CSS resources
CSS = OrderedDict((
    ('bootstrap', CSS_FILES[0]),
    ('fancybox', CSS_FILES[1]),
    ('datepicker', (
        '//cdnjs.cloudflare.com/ajax/libs/bootstrap-datepicker/'
        '1.6.0/css/bootstrap-datepicker.min.css')),
    ('bootstrap-ligo', CSS_FILES[2]),
    ('gwdetchar', CSS_FILES[3]),
    ('gwsumm', os.path.join(STATICDIR, 'gwsumm.min.css')),
))


# -- utilities ----------------------------------------------------------------

def get_css():
    """Return a `dict` of CSS files to link in the HTML <head>
    """
    return CSS


def get_js():
    """Return a `dict` of javascript files to link in the HTML <head>
    """
    return JS



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013-2015)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Utilies for GWSumm plotting
"""

import hashlib
import itertools
import re

from matplotlib import rcParams

from gwpy.plot.tex import label_to_latex
from gwpy.plot.utils import (  # noqa: F401
    FIGURE_PARAMS,
    AXES_PARAMS,
)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# -- plotting parameters ------------------------------------------------------

LINE_PARAMS = [
    'linewidth', 'linestyle', 'color', 'label', 'alpha', 'rasterized',
    'zorder',
]
COLLECTION_PARAMS = [
    'cmap', 'vmin', 'vmax', 'marker', 's', 'norm', 'rasterized',
]
IMAGE_PARAMS = [
    'imshow', 'cmap', 'vmin', 'vmax', 'norm', 'rasterized', 'extent',
    'origin', 'interpolation', 'aspect',
]
HIST_PARAMS = [
    'bins', 'range', 'normed', 'weights', 'cumulative', 'bottom',
    'histtype', 'align', 'orientation', 'rwidth', 'log', 'color',
    'label', 'stacked', 'logbins',
]
LEGEND_PARAMS = [
    'loc', 'borderaxespad', 'ncol',
]
ARTIST_PARAMS = set(itertools.chain.from_iterable([
    LINE_PARAMS,
    COLLECTION_PARAMS,
    IMAGE_PARAMS,
    HIST_PARAMS,
]))

# -- default labels for table columns -----------------------------------------

COLUMN_LABEL = {
    'peal_frequency': r"Frequency [Hz]",
    'central_freq': r"Frequency [Hz]",
    'frequency': r"Frequency [Hz]",
    'mchirp': r"Chirp mass [M$_\odot$]",
    'new_snr': r"$\chi^2$-weighted signal-to-noise ratio (New SNR)",
    'peak_frequency': r"Frequency [Hz]",
    'rho': r"$\rho$",
    'snr': r"Signal-to-noise ratio (SNR)",
    'template_duration': r"Template duration [s]",
}


def get_column_label(column):
    try:
        return COLUMN_LABEL[column]
    except KeyError:
        return get_column_string(column)


def get_column_string(column):
    r"""
    Format the string columnName (e.g. xml table column) into latex format for
    an axis label.

    Parameters
    ----------
    column : `str`
        string to format

    Examples
    --------
    >>> get_column_string('snr')
    'SNR'
    >>> get_column_string('bank_chisq_dof')
    r'Bank $\chi^2$ DOF'
    """
    acro = ['snr', 'ra', 'dof', 'id', 'ms', 'far']
    greek = ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'eta',
             'theta', 'iota', 'kappa', 'lamda', 'mu', 'nu', 'xi', 'omicron',
             'pi', 'rho', 'sigma', 'tau', 'upsilon', 'phi', 'chi', 'psi',
             'omega']
    unit = ['ns']
    sub = ['flow', 'fhigh', 'hrss', 'mtotal', 'mchirp']

    tex = rcParams['text.usetex']

    words = []
    for word in re.split(r'\s', column):
        if word.isupper():
            words.append(word)
        else:
            words.extend(re.split(r'_', word))

    # parse words
    for i, word in enumerate(words):
        # get acronym in lower case
        if word in acro:
            words[i] = word.upper()
        # get numerical unit
        elif word in unit:
            words[i] = '$(%s)$' % word
        # get character with subscript text
        elif word in sub and tex:
            words[i] = r'%s$_{\mbox{\small %s}}$' % (word[0], word[1:])
        # get greek word
        elif word in greek and tex:
            words[i] = r'$\%s$' % word
        # get starting with greek word
        elif re.match(r'(%s)' % '|'.join(greek), word) and tex:
            if word[-1].isdigit():
                words[i] = r'$\%s_{%s}$''' % tuple(
                    re.findall(r"[a-zA-Z]+|\d+", word))
            elif word.endswith('sq'):
                words[i] = r'$\%s^2$' % word.rstrip('sq')
        # get everything else
        else:
            if word.isupper():
                words[i] = word
            else:
                words[i] = word.title()
            # escape underscore
            words[i] = usetex_tex(re.sub(r'(?<!\\)_', r'\_', words[i]))
    return ' '.join(words)


def usetex_tex(text):
    """Format text for TeX if `text.usetex` is True
    """
    if rcParams['text.usetex']:
        return label_to_latex(text)
    return text


def hash(string, num=6):
    """Generate an N-character hash string based using string to initialise

    Parameters
    ----------
    string : `str`
        the initialisation string

    num : `int`, optional
        the length of the hash to produce

    Returns
    -------
    hash : `str`
        the new hash

    Examples
    --------
    >>> from gwsumm.plot.utils import hash
    >>> print(hash("I love gravitational waves"))
    80c897
    """
    return hashlib.md5(string.encode("utf-8")).hexdigest()[:num]



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Definitions for the standard plots
"""

from __future__ import division

import os.path
import warnings
from itertools import cycle

from six import string_types

import numpy

from matplotlib.colors import LogNorm

from astropy.units import Quantity

from gwpy.plot.colors import tint
from gwpy.plot.gps import GPSTransform
from gwpy.segments import SegmentList

from .. import (globalv, io)
from ..mode import (Mode, get_mode)
from ..utils import re_cchar
from ..data import (get_timeseries, get_spectrogram,
                    get_coherence_spectrogram, get_spectrum,
                    get_coherence_spectrum)
from ..state import ALLSTATE
from .registry import (get_plot, register_plot)
from .mixins import DataLabelSvgMixin
from .utils import usetex_tex

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

DataPlot = get_plot('data')
GREEN = (0.2, 0.8, 0.2)


class TimeSeriesDataPlot(DataLabelSvgMixin, DataPlot):
    """DataPlot of some `TimeSeries` data.
    """
    type = 'timeseries'
    data = 'timeseries'
    defaults = DataPlot.defaults.copy()
    defaults.update({
        'xscale': 'auto-gps',
        'yscale': 'linear',
    })

    def __init__(self, *args, **kwargs):
        super(TimeSeriesDataPlot, self).__init__(*args, **kwargs)
        if self.data == 'timeseries':
            for c in self.channels:
                c._timeseries = True

    def _update_defaults_from_channels(self):
        for chan in self.channels:
            if getattr(chan, 'amplitude_range', None) is not None:
                self.pargs.setdefault('ylim', chan.amplitude_range)
                break

    # -- utilities ------------------------------

    def add_state_segments(self, ax, visible=None, **kwargs):
        """Add an `Axes` below the given ``ax`` displaying the `SummaryState`
        for this `TimeSeriesDataPlot`.

        Parameters
        ----------
        ax : `Axes`
            the set of `Axes` below which to display the state segments.

        visible : `bool`, optional
            whether or not to display the axes, or just make space for them,
            default is `None`, meaning a dynamic choice based on the state

        **kwargs
            other keyword arguments will be passed to the
            :meth:`~gwpy.plot.Plot.add_segments_bar`
            method.
        """
        # allow user to disable the state segments axes
        if self.pargs.pop('no-state-segments', False):
            visible = False
        if visible is None and self.state is not None and (
                self.state.name.lower() != ALLSTATE):
            visible = True
        if visible:
            sax = self.plot.add_segments_bar(self.state, ax, height=.14,
                                             pad=.1,  **kwargs)
            sax.tick_params(axis='y', which='major', labelsize=12)
            sax.yaxis.set_ticks_position('none')
            sax.set_ylim(-.4, .4)
            return sax
        else:
            self.plot.subplots_adjust(bottom=0.17)
            return None

    def add_future_shade(self, gps=None, facecolor='gray', alpha=.1,
                         **kwargs):
        """Shade those parts of the figure that display times in the future
        """
        end = float(self.end)
        # get time 'now'
        if gps is None:
            gps = globalv.NOW
        # allow user to override
        if self.pargs.pop('no-future-shade', False) or end <= gps:
            return
        # shade time axes
        for ax in filter(
                lambda ax: isinstance(ax.xaxis.get_transform(), GPSTransform),
                self.plot.axes):
            ax.axvspan(gps, end, facecolor=facecolor, alpha=alpha, **kwargs)

    # -- init/finalize --------------------------

    def init_plot(self, *args, **kwargs):
        """Initialise the Figure and Axes objects for this
        `TimeSeriesDataPlot`.
        """
        epoch = kwargs.pop('epoch', self.pargs.pop('epoch', None))
        plot = super(TimeSeriesDataPlot, self).init_plot(*args, **kwargs)
        for ax in plot.axes:
            if get_mode() == Mode.month:
                ax.set_xscale('days')
            if isinstance(ax.xaxis.get_transform(), GPSTransform):
                ax.set_epoch(float(epoch if epoch is not None else self.start))
                if ax.get_autoscalex_on():
                    ax.set_xlim(float(self.start), float(self.end))
            ax.grid(True, which='both')
        return plot

    # -- main draw method -----------------------

    def draw(self, outputfile=None):
        """Read in all necessary data, and generate the figure.
        """
        plot = self.init_plot()
        ax = plot.gca()

        plotargs = self.parse_plot_kwargs()
        legendargs = self.parse_legend_kwargs()

        # add data
        channels, groups = list(zip(*self.get_channel_groups()))
        for clist, pargs in list(zip(groups, plotargs)):
            # get data
            valid = self._get_data_segments(clist[0])
            data = [get_timeseries(c, valid, query=False)
                    for c in clist]

            if len(clist) > 1:
                data = [tsl.join(gap='pad', pad=numpy.nan) for tsl in data]
            flatdata = [ts for tsl in data for ts in tsl]
            # validate parameters
            for ts in flatdata:
                # double-check empty
                if ts.x0 is None:
                    ts.epoch = self.start
                # double-check log scales
                if self.logy:
                    ts.value[ts.value == 0] = 1e-100
            # set label
            try:
                label = pargs.pop('label')
            except KeyError:
                try:
                    label = usetex_tex(flatdata[0].name)
                except IndexError:
                    label = clist[0]
                else:
                    if self.fileformat == 'svg' and not label.startswith(
                            usetex_tex(
                            str(flatdata[0].channel)).split('.')[0]):
                        label += ' [%s]' % (
                            usetex_tex(str(flatdata[0].channel)))

            # plot groups or single TimeSeries
            if len(clist) > 1:
                data[1].name = None  # force no labels for shades
                data[2].name = None
                ax.plot_mmm(*data, label=label, **pargs)
            elif len(flatdata) == 0:
                ax.plot(data[0].EntryClass([], epoch=self.start, unit='s',
                                           name=label),
                        label=label, **pargs)
            else:
                for ts in data[0]:
                    line, = ax.plot(ts, label=label, **pargs)
                    label = None
                    pargs['color'] = line.get_color()

        # customise plot
        self.add_hvlines()
        self.apply_parameters(ax, **self.pargs)

        # add legend
        if ax.get_legend_handles_labels()[0]:
            ax.legend(**legendargs)

        self.add_state_segments(ax)
        self.add_future_shade()

        return self.finalize(outputfile=outputfile)

    def _get_data_segments(self, channel):
        """Get data segments for this plot
        """
        if self.state and not self.all_data:
            return self.state.active
        if channel.sample_rate is not None:
            return SegmentList([self.span.protract(
                1/channel.sample_rate.value)])
        return SegmentList([self.span])


register_plot(TimeSeriesDataPlot)


class SpectrogramDataPlot(TimeSeriesDataPlot):
    """DataPlot a Spectrogram
    """
    type = 'spectrogram'
    data = 'spectrogram'
    defaults = TimeSeriesDataPlot.defaults.copy()
    defaults.update({
        'yscale': 'log',
        'ylabel': 'Frequency [Hz]',
        'ratio': None,
        'format': None,
        'rasterized': True,
    })

    def __init__(self, *args, **kwargs):
        super(SpectrogramDataPlot, self).__init__(*args, **kwargs)
        self.ratio = self.pargs.pop('ratio')
        # set default colour-map for median ratio
        if self.ratio in ('median', 'mean'):
            self.pargs.setdefault('cmap', 'Spectral_r')

    def _update_defaults_from_channels(self):
        for channel in self.channels:
            self.pargs.setdefault('ylim', channel.frequency_range)
            if isinstance(self.pargs['ylim'], Quantity):
                self.pargs['ylim'] = self.pargs['ylim'].value

            if self.ratio is None and self.pargs.get('clim') is None:
                if (self.pargs.get('format') in ('amplitude', 'asd') and
                        hasattr(channel, 'asd_range')):
                    self.pargs['clim'] = channel.asd_range
                elif hasattr(channel, 'psd_range'):
                    self.pargs['clim'] = channel.psd_range

    @property
    def pid(self):
        try:
            return self._pid
        except AttributeError:
            super(SpectrogramDataPlot, self).pid
            if (isinstance(self.ratio, string_types) and
                    os.path.isfile(self.ratio)):
                self._pid += '_REFERENCE_RATIO'
            elif self.ratio:
                self._pid += '_%s_RATIO' % re_cchar.sub(
                    '_', str(self.ratio).upper())
            return self.pid

    @pid.setter
    def pid(self, id_):
        self._pid = str(id_)

    @pid.deleter
    def pid(self):
        del self._pid

    def get_ratio(self, specgrams):
        ratio = self.ratio
        # calculate ratio spectrum
        if len(specgrams) and (
                ratio in ['median', 'mean'] or isinstance(ratio, int)):
            try:
                allspec = specgrams.join(gap='ignore')
            except ValueError as e:
                if 'units do not match' in str(e):
                    warnings.warn(str(e))
                    for spec in specgrams[1:]:
                        spec.unit = specgrams[0].unit
                    allspec = specgrams.join(gap='ignore')
                else:
                    raise
            if isinstance(ratio, int):
                return allspec.percentile(ratio)
            else:
                return getattr(allspec, ratio)(axis=0)
        elif isinstance(ratio, string_types) and os.path.isfile(ratio):
            try:
                return io.read_frequencyseries(ratio)
            except IOError as e:  # skip if file can't be read
                warnings.warn('IOError: %s' % str(e))
        return ratio

    def draw(self):
        # initialise
        plot = self.init_plot()
        ax = plot.gca()
        ax.grid(b=True, axis='y', which='major')
        channel = self.channels[0]

        # parse data arguments
        sdform = self.pargs.pop('format')

        # parse colorbar arguments
        clabel = self.pargs.pop('colorlabel', '')
        clim = self.pargs.pop('clim', None)
        clog = self.pargs.pop('logcolor', False)

        # parse plotting arguments
        if clim:  # clim -> (vmin, vmax)
            vmin, vmax = clim
            self.pargs.setdefault('vmin', vmin)
            self.pargs.setdefault('vmax', vmax)
        if clog:  # logcolor -> norm
            self.pargs.setdefault('norm', 'log')
        plotargs = self.parse_plot_kwargs()[0]  # only one channel

        # rework norm='log' into a LogNorm object
        # (this is only 'required' in the case of no data, when ax.scatter
        #  gets called manually below)
        if plotargs.get('norm', None) == 'log':
            vmin = self.pargs.get('vmin')
            vmax = self.pargs.get('vmax')
            plotargs['norm'] = LogNorm(vmin=vmin, vmax=vmax)

        # get data
        if self.state and not self.all_data:
            valid = self.state.active
        else:
            valid = SegmentList([self.span])
        if self.type == 'coherence-spectrogram':
            specgrams = get_coherence_spectrogram(self.channels, valid,
                                                  query=False)
        else:
            try:
                specgrams = get_spectrogram(channel, valid, query=False,
                                            format=sdform)
            except ValueError as exc:
                if 'need more than 0 values' not in str(exc):
                    raise
                # attempted to do math but one input has zero size
                specgrams = []

        # get ratio as FrequencySeries
        ratio = self.get_ratio(specgrams)

        # plot data
        for i, specgram in enumerate(specgrams):

            # calculate ratio
            if ratio is not None:
                specgram = specgram.ratio(ratio)

            # undo demodulation and crop frequencies
            ylim = self.pargs.get('ylim', None)
            specgram = undo_demodulation(specgram, channel, ylim)
            if ylim is not None:
                specgram = specgram.crop_frequencies(*ylim)

            # plot
            ax.imshow(specgram, **plotargs)

        # add colorbar
        if len(specgrams) == 0:
            ax.pcolormesh([1, 10], [1, 10], [[1, 10]], visible=False,
                          **plotargs)
        ax.colorbar(label=clabel)

        # customise and finalise
        self.apply_parameters(ax, **self.pargs)
        self.add_state_segments(ax)
        self.add_future_shade()

        return self.finalize()


register_plot(SpectrogramDataPlot)


class CoherenceSpectrogramDataPlot(SpectrogramDataPlot):
    """DataPlot a Spectrogram of the coherence between two channels
    """
    type = 'coherence-spectrogram'
    data = 'coherence-spectrogram'
    defaults = SpectrogramDataPlot.defaults.copy()
    defaults.update({
        'ratio': None,
        'format': None,
        'clim': None,
        'logcolor': False,
        'colorlabel': None,
    })


register_plot(CoherenceSpectrogramDataPlot)


class SpectrumDataPlot(DataPlot):
    """Spectrum plot for a `SummaryTab`
    """
    type = 'spectrum'
    data = 'spectrum'
    defaults = DataPlot.defaults.copy()
    defaults.update({
        'xscale': 'log',
        'yscale': 'log',
        'format': None,
        'zorder': 1,
        'no-percentiles': False,
        'reference-linestyle': '--',
    })

    def _update_defaults_from_channels(self):
        for channel in self.channels:
            if getattr(channel, 'frequency_range', None) is not None:
                self.pargs.setdefault('xlim', channel.frequency_range)
                if isinstance(self.pargs['xlim'], Quantity):
                    self.pargs['xlim'] = self.pargs['xlim'].value
            if (self.pargs.get('format') in ['amplitude', 'asd'] and
                    hasattr(channel, 'asd_range')):
                self.pargs.setdefault('ylim', channel.asd_range)
            elif hasattr(channel, 'psd_range'):
                self.pargs.setdefault('ylim', channel.psd_range)

    def draw(self):
        pargs = self.pargs.copy()
        try:
            self._draw()
        except OverflowError:
            self.pargs = pargs
            self.pargs['alpha'] = 0.0
            self._draw()

    def _draw(self):
        """Load all data, and generate this `SpectrumDataPlot`
        """
        plot = self.init_plot()
        ax = plot.gca()
        ax.grid(b=True, axis='both', which='both')

        if self.state:
            self.pargs.setdefault(
                'suptitle',
                '[%s-%s, state: %s]' % (self.span[0], self.span[1],
                                        usetex_tex(str(self.state))))
        suptitle = self.pargs.pop('suptitle', None)
        if suptitle:
            plot.suptitle(suptitle, y=0.993, va='top')

        # get spectrum format: 'amplitude' or 'power'
        sdform = self.pargs.pop('format')
        if sdform == 'rayleigh':
            method = 'rayleigh'
        else:
            method = None
        use_percentiles = str(
            self.pargs.pop('no-percentiles')).lower() == 'false'

        # parse plotting arguments
        plotargs = self.parse_plot_kwargs()
        legendargs = self.parse_legend_kwargs()
        use_legend = False

        # get reference arguments
        refs = self.parse_references()

        # add data
        if self.type == 'coherence-spectrum':
            iterator = list(zip(self.channels[0::2], self.channels[1::2],
                                plotargs))
        else:
            iterator = list(zip(self.channels, plotargs))

        for chantuple in iterator:
            channel = chantuple[0]
            channel2 = chantuple[1]
            pargs = chantuple[-1]

            if self.state and not self.all_data:
                valid = self.state
            else:
                valid = SegmentList([self.span])

            if self.type == 'coherence-spectrum':
                data = get_coherence_spectrum([str(channel), str(channel2)],
                                              valid, query=False)
            else:
                try:
                    data = get_spectrum(str(channel), valid, query=False,
                                        format=sdform, method=method)
                except ValueError as exc:
                    # math op failed beacuse one of the datasets is empty
                    if (
                            'could not be broadcast' in str(exc) and
                            '(0,)' in str(exc)
                    ):
                        data = []
                    else:
                        raise

            # undo demodulation
            data = list(data)
            for i, spec in enumerate(data):
                data[i] = undo_demodulation(spec, channel,
                                            self.pargs.get('xlim', None))

            # anticipate log problems
            if self.logx:
                data = [s[1:] for s in data]
            if self.logy:
                for sp in data:
                    sp.value[sp.value == 0] = 1e-100

            if 'label' in pargs:
                use_legend = True

            if data and use_percentiles:
                _, minline, maxline, _ = ax.plot_mmm(*data, **pargs)
                # make min, max lines lighter:
                minline.set_alpha(pargs.get('alpha', .1) * 2)
                maxline.set_alpha(pargs.get('alpha', .1) * 2)
            elif data:
                ax.plot(data[0], **pargs)

        # display references
        for source, refparams in refs.items():
            refspec = io.read_frequencyseries(source)
            refparams.setdefault('zorder', -len(refs) + 1)
            if 'filter' in refparams:
                refspec = refspec.filter(*refparams.pop('filter'))
            if 'scale' in refparams:
                refspec *= refparams.pop('scale', 1)
            if 'label' in refparams:
                use_legend = True
            ax.plot(refspec, **refparams)

        # customise
        self.add_hvlines()
        self.apply_parameters(ax, **self.pargs)
        if use_legend or len(self.channels) > 1 or ax.legend_ is not None:
            ax.legend(**legendargs)

        return self.finalize()

    def parse_references(self, prefix=r'reference(\d+)?\Z'):
        """Parse parameters for displaying one or more reference traces
        """
        return self.parse_list('reference')


register_plot(SpectrumDataPlot)


class CoherenceSpectrumDataPlot(SpectrumDataPlot):
    """Coherence pectrum plot for a `SummaryTab`
    """
    type = 'coherence-spectrum'
    data = 'coherence-spectrogram'
    defaults = SpectrumDataPlot.defaults.copy()
    defaults.update({
        'yscale': 'linear',
        'alpha': 0.1,
    })

    # override this to allow us to set the legend manually
    def _parse_labels(self, defaults=None):
        return self.pargs.pop('labels', defaults)

    def get_channel_groups(self):
        """Hi-jacked method to return pairs of channels

        For the `CoherenceSpectrumDataPlot` this method is only used in
        determining how to separate lists of plotting argument given by
        the user.
        """
        all_ = self.allchannels
        return [(all_[i], all_[i:i+2]) for i in range(0, len(all_), 2)]


register_plot(CoherenceSpectrumDataPlot)


class TimeSeriesHistogramPlot(DataPlot):
    """HistogramPlot from a Series
    """
    type = 'histogram'
    data = 'timeseries'
    defaults = DataPlot.defaults.copy()
    defaults.update({
        'ylabel': 'Rate [Hz]',
        'log': True,
        'histtype': 'stepfilled',
        'rwidth': 1,
        'bottom': 1e-300,
    })

    def _update_defaults_from_channels(self):
        for channel in self.channels:
            if hasattr(channel, 'amplitude_range'):
                self.pargs.setdefault('xlim', channel.amplitude_range)
                break

    def init_plot(self, geometry=None, **kwargs):
        """Initialise the Figure and Axes objects for this
        `TimeSeriesDataPlot`.
        """
        if geometry is None and self.pargs.pop('sep', False):
            if self.data == 'segments':
                geometry = (len(self.flags), 1)
            else:
                geometry = (len(self.channels), 1)
        elif geometry is None:
            geometry = (1, 1)
        plot = super(TimeSeriesHistogramPlot, self).init_plot(
            geometry=geometry, **kwargs)
        for ax in plot.axes:
            ax.grid(True, which='both')
        return plot

    def parse_plot_kwargs(self, **defaults):
        kwargs = super(TimeSeriesHistogramPlot, self).parse_plot_kwargs(
            **defaults)
        for histargs in kwargs:
            histargs.setdefault('logbins', self.logx)
            logy = histargs.get('log', False)
            self.pargs.setdefault('yscale', 'log' if logy else 'linear')
            # set range as xlim
            if 'range' not in histargs and 'xlim' in self.pargs:
                histargs['range'] = self.pargs.get('xlim')
            # set alpha
            if len(self.channels) > 1:
                histargs.setdefault('alpha', 0.7)
        return kwargs

    def draw(self, outputfile=None):
        """Get data and generate the figure.
        """
        # get plot and axes
        plot = self.init_plot()
        axes = plot.axes

        if self.state:
            self.pargs.setdefault(
                'suptitle',
                '[%s-%s, state: %s]' % (self.span[0], self.span[1],
                                        usetex_tex(str(self.state))))
        suptitle = self.pargs.pop('suptitle', None)
        if suptitle:
            plot.suptitle(suptitle, y=0.993, va='top')

        # extract histogram arguments
        histargs = self.parse_plot_kwargs()

        # get data
        data = []
        for channel in self.channels:
            if self.state and not self.all_data:
                valid = self.state.active
            else:
                valid = SegmentList([self.span])
            data.append(get_timeseries(channel, valid, query=False).join(
                gap='ignore', pad=numpy.nan))

        # plot
        for ax, arr, pargs in zip(cycle(axes), data, histargs):
            # set range if not given
            if pargs.get('range') is None:
                pargs['range'] = self._get_range(
                    data,
                    # use range from first dataset if already calculated
                    range=histargs[0].get('range'),
                    # use xlim if manually set (user or INI)
                    xlim=None if ax.get_autoscalex_on() else ax.get_xlim(),
                )

            # plot histogram
            _, _, patches = ax.hist(arr, **pargs)

            # update edge color of histogram to be tinted version of face
            if pargs.get('histtype', None) == 'stepfilled':
                for p in patches:
                    if not p.get_edgecolor()[3]:
                        p.set_edgecolor(tint(p.get_facecolor(), .7))

        # customise plot
        legendargs = self.parse_legend_kwargs()
        for i, ax in enumerate(axes):
            for key, val in self.pargs.items():
                if key == 'title' and i > 0:
                    continue
                if key == 'xlabel' and i < (len(axes) - 1):
                    continue
                if key == 'ylabel' and (
                        (len(axes) % 2 and i != len(axes) // 2) or
                        (len(axes) % 2 == 0 and i > 0)):
                    continue
                try:
                    getattr(ax, 'set_%s' % key)(val)
                except AttributeError:
                    setattr(ax, key, val)
            if len(self.channels) > 1:
                ax.legend(**legendargs)
        if len(axes) % 2 == 0 and axes[0].get_ylabel():
            label = axes[0].yaxis.label
            ax = axes[int(len(axes) // 2)-1]
            ax.set_ylabel(label.get_text())
            ax.yaxis.label.set_position((0, -.2 / len(axes)))
            if len(axes) != 2:
                label.set_text('')

        # add extra axes and finalise
        return self.finalize(outputfile=outputfile)

    def _get_range(self, data, range=None, xlim=None):
        if range is not None or xlim is not None:
            return range or xlim
        try:
            return numpy.min(data), numpy.max(data)
        except ValueError as exc:
            if not str(exc).startswith('zero-size array'):
                raise
        return None


register_plot(TimeSeriesHistogramPlot)


class TimeSeriesHistogram2dDataPlot(TimeSeriesHistogramPlot):
    """DataPlot of the 2D histogram of two `TimeSeries`.
    """
    type = 'histogram2d'
    data = 'timeseries'
    defaults = TimeSeriesHistogramPlot.defaults.copy()
    defaults.update({
        'yscale': 'linear',
        'grid': 'both',
        'shading': 'flat',
        'cmap': 'inferno_r',
        'alpha': None,
        'edgecolors': 'None',
        'bins': 100,
        'normed': True
    })

    def __init__(self, *args, **kwargs):
        super(TimeSeriesHistogram2dDataPlot, self).__init__(*args, **kwargs)
        channels = self.channels
        if isinstance(channels, (list, tuple)) and len(channels) > 2:
            raise ValueError("Cannot generate TimeSeriesHistogram2dDataPlot "
                             " plot with more than 2 channels")

    def _update_defaults_from_channels(self):
        c1, c2 = self.channels
        self.pargs.setdefault('xlabel', usetex_tex(str(c1)))
        self.pargs.setdefault('ylabel', usetex_tex(str(c2)))
        if hasattr(c1, 'amplitude_range'):
            self.pargs.setdefault('xlim', c1.amplitude_range)
        if hasattr(c2, 'amplitude_range'):
            self.pargs.setdefault('ylim', c2.amplitude_range)

    def parse_hist_kwargs(self, **defaults):
        kwargs = {'bins': self.pargs.pop('bins'),
                  'normed': self.pargs.pop('normed')}
        if 'range' in self.pargs:
            ranges = [float(r) for r in self.pargs['range'].split(',')]
            kwargs['range'] = [[ranges[0], ranges[1]],
                               [ranges[2], ranges[3]]]
        elif 'xlim' in self.pargs and 'ylim' in self.pargs:
            xlim = self.pargs['xlim']
            ylim = self.pargs['ylim']
            kwargs['range'] = [[xlim[0], xlim[1]], [ylim[0], ylim[1]]]
        else:
            kwargs['range'] = None
        return kwargs

    def parse_pcmesh_kwargs(self, **defaults):
        kwargs = {
                  'cmap': self.pargs.pop('cmap'),
                  'edgecolors': self.pargs.pop('edgecolors'),
                  'shading': self.pargs.pop('shading'),
                  'alpha': self.pargs.pop('alpha')
                 }
        return kwargs

    def draw(self, outputfile=None):
        """Get data and generate the figure.
        """
        # get histogram parameters
        plot = self.init_plot()
        ax = plot.gca()

        if self.state:
            self.pargs.setdefault(
                'suptitle',
                '[%s-%s, state: %s]' % (self.span[0], self.span[1],
                                        usetex_tex(str(self.state))))
        suptitle = self.pargs.pop('suptitle', None)
        if suptitle:
            plot.suptitle(suptitle, y=0.993, va='top')
        # get data
        data = []
        for channel in self.channels:
            if self.state and not self.all_data:
                valid = self.state.active
            else:
                valid = SegmentList([self.span])
            data.append(get_timeseries(channel, valid, query=False).join(
                gap='ignore', pad=numpy.nan))
        if len(data) == 1:
            data.append(data[0])
        # histogram
        hist_kwargs = self.parse_hist_kwargs()
        h, xedges, yedges = numpy.histogram2d(data[0], data[1], **hist_kwargs)
        h = numpy.ma.masked_where(h == 0, h)
        x, y = numpy.meshgrid(xedges, yedges, copy=False, sparse=True)
        # plot
        pcmesh_kwargs = self.parse_pcmesh_kwargs()
        ax.pcolormesh(x, y, h.T, **pcmesh_kwargs)

        # customise plot
        self.apply_parameters(ax, **self.pargs)

        return self.finalize(outputfile=outputfile)


register_plot(TimeSeriesHistogram2dDataPlot)


class SpectralVarianceDataPlot(SpectrumDataPlot):
    """SpectralVariance histogram plot for a `DataTab`
    """
    type = 'variance'
    data = 'spectrogram'
    defaults = SpectrumDataPlot.defaults.copy()
    defaults.update({
        'xscale': 'log',
        'yscale': 'log',
        'reference-linestyle': '--',
        'log': True,
        'nbins': 100,
    })

    def __init__(self, channels, *args, **kwargs):
        if isinstance(channels, (list, tuple)) and len(channels) > 1:
            raise ValueError("Cannot generate SpectralVariance plot with "
                             "more than 1 channel")
        super(SpectralVarianceDataPlot, self).__init__(
            channels, *args, **kwargs)

    def _update_defaults_from_channels(self):
        chan = self.channels[0]

        if getattr(chan, 'frequency_range', None) is not None:
            self.pargs.setdefault('xlim', chan.frequency_range)
            if isinstance(self.pargs['xlim'], Quantity):
                self.pargs['xlim'] = self.pargs['xlim'].value

        if hasattr(chan, 'asd_range'):
            self.pargs.setdefault('ylim', chan.asd_range)

        if hasattr(chan, 'asd_range'):
            low, high = chan.asd_range
            self.pargs.setdefault('low', low)
            self.pargs.setdefault('high', high)

    def parse_variance_kwargs(self):
        varargs = dict()
        for key in ['low', 'high', 'log', 'nbins', 'bins', 'density', 'norm']:
            if key in self.pargs:
                varargs[key] = self.pargs.pop(key)
        return varargs

    def _draw(self):
        """Load all data, and generate this `SpectrumDataPlot`
        """
        plot = self.init_plot()
        ax = plot.gca()

        if self.state:
            self.pargs.setdefault(
                'suptitle',
                '[%s-%s, state: %s]' % (self.span[0], self.span[1],
                                        usetex_tex(str(self.state))))
        suptitle = self.pargs.pop('suptitle', None)
        if suptitle:
            plot.suptitle(suptitle, y=0.993, va='top')

        # parse plotting arguments
        cmap = self.pargs.pop('cmap', None)
        varargs = self.parse_variance_kwargs()
        plotargs = self.parse_plot_kwargs()[0]

        # get reference arguments
        refs = self.parse_references()

        # calculate spectral variance and plot
        # pad data request to over-fill plots (no gaps at the end)
        if self.state and not self.all_data:
            valid = self.state.active
        else:
            valid = SegmentList([self.span])
        livetime = float(abs(valid))

        if livetime:
            plotargs.setdefault('vmin', 1/livetime)
        plotargs.setdefault('vmax', 1.)
        plotargs.pop('label')

        specgram = get_spectrogram(self.channels[0], valid, query=False,
                                   format='asd').join(gap='ignore')

        if specgram.size:
            asd = specgram.median(axis=0)
            asd.name = None
            variance = specgram.variance(**varargs)
            # normalize the variance
            variance /= livetime / specgram.dt.value
            # undo demodulation
            variance = undo_demodulation(variance, self.channels[0],
                                         ax.get_xlim())

            # plot
            ax.plot(asd, color='grey', linewidth=0.3)
            ax.imshow(variance, cmap=cmap, **plotargs)

        # display references
        for source, refparams in refs.items():
            refspec = io.read_frequencyseries(source)
            if 'filter' in refparams:
                refspec = refspec.filter(*refparams.pop('filter'))
            if 'scale' in refparams:
                refspec *= refparams.pop('scale', 1)
            ax.plot(refspec, **refparams)

        # customise
        self.add_hvlines()
        self.apply_parameters(ax, **self.pargs)
        ax.grid(b=True, axis='both', which='both')

        return self.finalize()


register_plot(SpectralVarianceDataPlot)


class RayleighSpectrogramDataPlot(SpectrogramDataPlot):
    """Rayleigh statistic versino of `SpectrogramDataPlot`
    """
    type = 'rayleigh-spectrogram'
    data = 'rayleigh-spectrogram'
    defaults = SpectrogramDataPlot.defaults.copy()
    defaults.update({
        'ratio': None,
        'format': 'rayleigh',
        'clim': [0.25, 4],
        'cmap': 'BrBG_r',
        'colorlabel': 'Rayleigh statistic',
    })


register_plot(RayleighSpectrogramDataPlot)


class RayleighSpectrumDataPlot(SpectrumDataPlot):
    """Rayleigh statistic versino of `SpectrumDataPlot`
    """
    type = 'rayleigh-spectrum'
    data = 'rayleigh-spectrum'
    defaults = {'format': 'rayleigh',
                'xscale': 'log',
                'yscale': 'log',
                'alpha': 0.1,
                'zorder': 1,
                'no-percentiles': True,
                'reference-linestyle': '--'}


register_plot(RayleighSpectrumDataPlot)


def undo_demodulation(spec, channel, limits=None):
    if spec.size == 0:
        return spec
    # undo demodulation
    try:
        demod = channel.demodulation
    except AttributeError:
        return spec
    else:
        spec = spec[:]  # views data with copied metadata
        del spec.frequencies
        spec.f0 = demod
        # if physical frequency-range is below demod, get negative df
        try:
            low, high = channel.frequency_range
        except (AttributeError, TypeError):
            try:
                low, high = limits
            except TypeError:
                return spec
        high = Quantity(high, 'Hz')
        if high < spec.f0:
            if spec.ndim > 1:  # Spectrogram
                spec.value[:] = numpy.fliplr(spec.value)
            else:  # FrequencySeries
                spec.value[:] = spec.value[::-1]
            spec.df *= -1
            spec.frequencies = spec.frequencies[::-1]
        return spec



# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""Custom `SummaryTab` for the output of the FScan algorithm.
"""

import os
import re
import glob

from six import string_types

from .registry import (get_tab, register_tab)

from .. import html
from ..mode import Mode
from ..plot import get_plot
from ..config import GWSummConfigParser

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['StampPEMTab']

base = get_tab('default')
SummaryPlot = get_plot(None)


class StampPEMTab(base):
    """Custom tab displaying a summary of StampPEM results.
    """
    type = 'stamp'

    def __init__(self, *args, **kwargs):
        if kwargs['mode'] != Mode.day:
            raise RuntimeError("StampPEMTab is only available in %s mode."
                               % Mode.day.name)
        super(StampPEMTab, self).__init__(*args, **kwargs)

    @classmethod
    def from_ini(cls, config, section, **kwargs):
        """Define a new `StampPEMTab` from a `ConfigParser`.
        """
        # parse generic configuration
        new = super(StampPEMTab, cls).from_ini(config, section, **kwargs)
        new.set_layout([2])

        # work out day directory and url
        new.directory = os.path.normpath(config.get(section, 'base-directory'))
        try:
            home_, postbase = new.directory.split('/public_html/', 1)
        except ValueError as e:
            e.args = ('Stamp PEM directory not under \'public_html\', '
                      'cannot format linkable URL',)
            raise
        else:
            user = os.path.split(home_)[1]
            new.url = '/~%s/%s' % (user, postbase.rstrip('/'))
        return new

    def process(self, config=GWSummConfigParser(), **kwargs):
        # find all plots
        self.plots = []
        if isinstance(self.directory, string_types):
            plots = sorted(
                glob.glob(os.path.join(self.directory, 'DAY_*.png')),
                key=lambda p: float(re.split(r'[-_]', os.path.basename(p))[1]))
            for p in plots:
                pname = os.path.split(p)[1]
                self.plots.append(SummaryPlot(
                    src=os.path.join(self.url, pname),
                    href=os.path.join(self.url,
                                      pname.replace('.png', '.html'))))

    def write_state_html(self, state):
        """Write the '#main' HTML content for this `StampPEMTab`.
        """
        page = markup.page()

        a = markup.oneliner.a('analysis', href=self.url+'/',
                              class_='alert-link', rel='external',
                              target='_blank')
        if not(os.path.isdir(self.directory)):
            page.div(class_='alert alert-warning', role='alert')
            page.p("No %s was performed for this period, "
                   "please try again later." % a)
            page.p("If you believe these data should have been found, please "
                   "contact %s."
                   % markup.oneliner.a('the DetChar group',
                                       class_='alert-link',
                                       href='mailto:detchar@ligo.org'))
            page.div.close()

        elif not self.plots:
            page.div(class_='alert alert-warning', role='alert')
            page.p("This %s produced no plots." % a)
            page.p("If you believe these data should have been found, please "
                   "contact %s."
                   % markup.oneliner.a('the DetChar group',
                                       class_='alert-link',
                                       href='mailto:detchar@ligo.org'))
            page.div.close()

        else:
            page.add(str(self.scaffold_plots(
                aclass='fancybox-stamp plot',
                **{'data-fancybox-type': 'iframe'})))
            page.hr(class_='row-divider')

            # link full results
            page.hr(class_='row-divider')
            page.div(class_='btn-group')
            page.a('Click here for the full Stamp PEM results',
                   href=self.url+'/', rel='external', target='_blank',
                   class_='btn btn-default btn-info btn-xl')
            page.div.close()

        # write to file
        idx = self.states.index(state)
        with open(self.frames[idx], 'w') as fobj:
            fobj.write(str(page))
        return self.frames[idx]


register_tab(StampPEMTab)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""This module defines a number of `Tab` subclasses.

The `builtin` classes provide interfaces for simple operations including

- `ExternalTab`: embedding an existing webpage
- `PlotTab`: scaffolding a collection of existing images
- `StateTab`: scaffolding plots split into a number of states, with
  a button to switch between states in the HTML

"""

import os.path
import warnings
from configparser import NoOptionError

from six import string_types

from MarkupPy import markup

from .registry import (get_tab, register_tab)
from ..plot import get_plot
from ..utils import (re_quote, re_cchar)
from ..config import GWSummConfigParser
from ..state import (ALLSTATE, SummaryState, get_state)
from .. import html

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['ExternalTab', 'PlotTab', 'StateTab']

Tab = get_tab('basic')
SummaryPlot = get_plot(None)
DataPlot = get_plot('data')


class ExternalTab(Tab):
    """A simple tab to link HTML from an external source

    Parameters
    ----------
    name : `str`
        name of this tab (required)
    url : `str`
        URL of the external content to be linked into this tab.
    index : `str`
        HTML file in which to write. By default each tab is written to
        an index.html file in its own directory. Use :attr:`~Tab.index`
        to find out the default index, if not given.
    shortname : `str`
        shorter name for this tab to use in the navigation bar. By
        default the regular name is used
    parent : :class:`~gwsumm.tabs.Tab`
        parent of this tab. This is used to position this tab in the
        navigation bar.
    children : `list`
        list of child :class:`Tabs <~gwsumm.tabs.Tab>` of this one. This
        is used to position this tab in the navigation bar.
    group : `str`
        name of containing group for this tab in the navigation bar
        dropdown menu. This is only relevant if this tab has a parent.
    path : `str`
        base output directory for this tab (should be the same directory
        for all tabs in this run).

    Configuration
    -------------

    """
    type = 'external'

    def __init__(self, name, url, error=True, success=None, **kwargs):
        """Initialise a new `ExternalTab`.
        """
        super(ExternalTab, self).__init__(name, **kwargs)
        self.url = url
        self.error = error
        self.success = success

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, link):
        self._url = link

    @classmethod
    def from_ini(cls, cp, section, *args, **kwargs):
        """Configure a new `ExternalTab` from a `ConfigParser` section

        Parameters
        ----------
        cp : :class:`~gwsumm.config.ConfigParser`
            configuration to parse.
        section : `str`
            name of section to read

        See Also
        --------
        Tab.from_ini :
            for documentation of the standard configuration
            options

        Notes
        -----
        On top of the standard configuration options, the `ExternalTab` can
        be configured with the ``url`` option, specifying the URL of the
        external content to be included:

        .. code-block:: ini

           [tab-external]
           name = External data
           type = external
           url = https://www.example.org/index.html


        """
        url = cp.get(section, 'url')
        if cp.has_option(section, 'error'):
            kwargs.setdefault(
                'error', re_quote.sub('', cp.get(section, 'error')))
        if cp.has_option(section, 'success'):
            kwargs.setdefault(
                'success', re_quote.sub('', cp.get(section, 'success')))
        return super(ExternalTab, cls).from_ini(cp, section, url,
                                                *args, **kwargs)

    def html_content(self, content):
        wrappedcontent = html.load(self.url, id_='content', error=self.error,
                                   success=self.success)
        return super(ExternalTab, self).html_content(wrappedcontent)

    def write_html(self, **kwargs):
        """Write the HTML page for this tab.

        See Also
        --------
        gwsumm.tabs.Tab.write_html : for details of all valid keyword
        arguments
        """
        if not kwargs.pop('writehtml', True):
            return
        link = markup.given_oneliner.a('click here to view the original',
                                       class_='reference',
                                       href=self.url.split()[0])
        kwargs.setdefault('footer', 'This page contains data from an external '
                                    'source, %s.' % link)
        return super(ExternalTab, self).write_html('', **kwargs)


register_tab(ExternalTab)


class PlotTab(Tab):
    """A simple tab to layout some figures in the #main div.

    Parameters
    ----------
    name : `str`
        name of this tab (required)
    plots : `list`, optional
        list of plots to display on this tab. More plots can be added
        at any time via :meth:`PlotTab.add_plot`
    layout : `int`, `list`, optional
        the number of plots to display in each row, or a list of numbers
        to define each row individually. If the number of plots defined
        by the layout is less than the total number of plots, the layout
        for the final row will be repeated as necessary.

        For example ``layout=[1, 2, 3]`` will display a single plot on
        the top row, two plots on the second, and 3 plots on each row
        thereafter.
    foreword : `~MarkupPy.markup.page`, `str`, optional
        content to include in the #main HTML before the plots
    afterword : `~MarkupPy.markup.page`, `str`, optional
        content to include in the #main HTML after the plots
    index : `str`, optional
        HTML file in which to write. By default each tab is written to
        an index.html file in its own directory. Use :attr:`~Tab.index`
        to find out the default index, if not given.
    shortname : `str`, optional
        shorter name for this tab to use in the navigation bar. By
        default the regular name is used
    parent : :class:`~gwsumm.tabs.Tab`, optional
        parent of this tab. This is used to position this tab in the
        navigation bar.
    children : `list`, optional
        list of child :class:`Tabs <~gwsumm.tabs.Tab>` of this one. This
        is used to position this tab in the navigation bar.
    group : `str`, optional
        name of containing group for this tab in the navigation bar
        dropdown menu. This is only relevant if this tab has a parent.
    path : `str`, optional,
        base output directory for this tab (should be the same directory
        for all tabs in this run).
    """
    type = 'plots'

    def __init__(self, name, plots=list(), layout=None, foreword=None,
                 afterword=None, **kwargs):
        """Initialise a new :class:`PlotTab`.
        """
        super(PlotTab, self).__init__(name, **kwargs)
        self.plots = []
        for p in plots:
            self.add_plot(p)
        self.set_layout(layout)
        self.foreword = foreword
        self.afterword = afterword

    @property
    def layout(self):
        """List of how many plots to display on each row in the output.

        By default this is ``1`` if the tab contains only 1 or 3 plots,
        or ``2`` if otherwise.
        The final number given in the list will be repeated as necessary.

        :type: `list` of `ints <int>`
        """
        return self._layout

    @layout.setter
    def layout(self, layout_):
        warnings.warn("Use of `Tab.layout = ...` has been deprecated, "
                      "please switch to using `Tab.set_layout(...)`",
                      DeprecationWarning)
        self.set_layout(layout_)

    def set_layout(self, layout):
        """Set the plot scaffolding layout for this tab

        Parameters
        ----------
        l : `int`, `list` of `int`
            the desired scaffold layout, one of

            - an `int`, indicating the number of plots on every row, or
            - a `list`, indicating the number of plots on each row, with
              the final `int` repeated for any remaining rows; each entry
              should be an `int` or a pair of `int` indicating the
              number of plots on this row AND the desired the scaling of
              this row, see the examples...

        Examples
        --------
        To layout 2 plots on each row

        >>> tab.set_layout(2)

        or

        >>> tab.set_layout([2])

        To layout 2 plots on the first row, and 3 on all other rows

        >>> tab.set_layout((2, 3))

        To layout 2 plots on the first row, and 1 on the second row BUT
        have it the same size as plots on a 2-plot row

        >>> tab.set_layout((2, (1, 2))
        """
        # shortcut None
        if layout is None:
            self._layout = None
            return
        # parse a single int
        if isinstance(layout, int) or (
                isinstance(layout, string_types) and
                layout.isdigit()
        ):
            self._layout = [int(layout)]
            return
        # otherwise parse as a list of ints or pairs of ints
        if isinstance(layout, string_types):
            layout = layout.split(',')
        self._layout = []
        for item in layout:
            if isinstance(item, int):
                self._layout.append(item)
                continue
            if isinstance(item, string_types):
                item = item.strip('([').rstrip(')]').split(',')
            if not isinstance(item, (list, tuple)) or not len(item) == 2:
                raise ValueError("Cannot parse layout element %r (%s)"
                                 % (item, type(item)))
            self._layout.append(tuple(map(int, item)))

    @property
    def foreword(self):
        """HTML content to be included before the plots
        """
        return self._pre

    @foreword.setter
    def foreword(self, content):
        if isinstance(content, markup.page) or content is None:
            self._pre = content
        else:
            self._pre = markup.page()
            if not str(content).startswith('<'):
                self._pre.p(str(content))
            else:
                self._pre.add(str(content))

    @property
    def afterword(self):
        """HTML content to be included after the plots
        """
        return self._post

    @afterword.setter
    def afterword(self, content):
        if isinstance(content, markup.page) or content is None:
            self._post = content
        else:
            self._post = markup.page()
            if not str(content).startswith('<'):
                self._post.p(str(content))
            else:
                self._post.add(str(content))

    @classmethod
    def from_ini(cls, cp, section, *args, **kwargs):
        """Define a new tab from a :class:`~gwsumm.config.GWConfigParser`

        Parameters
        ----------
        cp : :class:`~ConfigParser.GWConfigParser`
            customised configuration parser containing given section
        section : `str`
            name of section to parse

        Returns
        -------
        tab : `PlotTab`
            a new tab defined from the configuration
        """
        cp = GWSummConfigParser.from_configparser(cp)

        kwargs.setdefault('path', '')

        # get layout
        if cp.has_option(section, 'layout'):
            try:
                layout = eval(cp.get(section, 'layout'))
            except NameError:
                raise ValueError("Cannot parse 'layout' for '%s' tab. Layout "
                                 "should be given as a comma-separated list "
                                 "of integers")
            if isinstance(layout, int):
                layout = [layout]
            for item in layout:
                if isinstance(item, (tuple, list)):
                    item = item[0]
                if item > 12:
                    raise ValueError("Cannot print more than 12 plots in a "
                                     "single row. The chosen layout value for "
                                     "each row must be a divisor of 12 to fit "
                                     "the Bootstrap scaffolding. For details "
                                     "see http://getbootstrap.com/2.3.2/"
                                     "scaffolding.html")
        else:
            layout = None
        kwargs.setdefault('layout', layout)

        # get plots
        if 'plots' not in kwargs:
            kwargs['plots'] = []
            for idx, url in sorted([(int(opt), url) for (opt, url) in
                                    cp.nditems(section) if opt.isdigit()],
                                   key=lambda x: x[0]):
                plot = SummaryPlot(href=url)
                plot.new = False  # this plot does not need to be generated
                # get caption
                try:
                    plot.caption = re_quote.sub(
                        '', cp.get(section, '%d-caption' % idx))
                except NoOptionError:
                    pass
                kwargs['plots'].append(plot)

        # get content
        try:
            kwargs.setdefault('foreword', cp.get(section, 'foreword'))
        except NoOptionError:
            pass
        try:
            kwargs.setdefault('afterword', cp.get(section, 'afterword'))
        except NoOptionError:
            pass

        # build and return tab
        return super(PlotTab, cls).from_ini(cp, section, *args, **kwargs)

    def add_plot(self, plot):
        """Add a plot to this tab.

        Parameters
        ----------
        plot : `str`, :class:`~gwsumm.plot.SummaryPlot`
            either the URL of a plot to embed, or a formatted `SummaryPlot`
            object.
        """
        if isinstance(plot, string_types):
            plot = SummaryPlot(href=plot)
            plot.new = False
        if not isinstance(plot, SummaryPlot):
            raise TypeError("Cannot append plot of type %r" % type(plot))
        self.plots.append(plot)

    def scaffold_plots(self, plots=None, state=None, layout=None,
                       aclass='fancybox plot', **fancyboxargs):
        """Build a grid of plots using bootstrap's scaffolding.

        Returns
        -------
        page : :class:`~MarkupPy.markup.page`
            formatted markup with grid of plots
        """
        page = markup.page()
        page.div(class_='scaffold well')

        if plots is None:
            if state:
                plots = [p for p in self.plots if not
                         isinstance(p, DataPlot) or p.state in [state, None]]
            else:
                plots = self.plots

        # get layout
        if layout is None:
            if self.layout:
                layout = list(self.layout)
            else:
                layout = len(plots) == 1 and [1] or [2]
        for i, item in enumerate(layout):
            if isinstance(item, (list, tuple)):
                layout[i] = item
            elif isinstance(item, int):
                layout[i] = (item, None)
            else:
                raise ValueError("Cannot parse layout element '%s'." % item)
        while sum(list(zip(*layout))[0]) < len(plots):
            layout.append(layout[-1])
        l = i = 0
        fancyboxargs.setdefault('data-fancybox-group', 1)

        for j, plot in enumerate(plots):
            # start new row
            if i == 0:
                page.div(class_='row')
            # determine relative size
            if layout[l][1]:
                colwidth = 12 // int(layout[l][1])
                remainder = 12 - colwidth * layout[l][0]
                if remainder % 2:
                    raise ValueError("Cannot center column of width %d in a "
                                     "12-column format" % colwidth)
                else:
                    offset = remainder / 2
                page.div(class_='col-md-%d col-md-offset-%d'
                                % (colwidth, offset))
            else:
                colwidth = 12 // int(layout[l][0])
                page.div(class_='col-md-%d' % colwidth)
            if plot.src.endswith('svg'):
                fbkw = fancyboxargs.copy()
                fbkw['data-fancybox-type'] = 'iframe'
                page.a(href='%s?iframe' % plot.href.replace('.svg', '.html'),
                       class_=aclass, **fbkw)
            else:
                fbkw = fancyboxargs.copy()
                fbkw['title'] = plot.caption
                page.a(href=plot.href, class_=aclass, **fbkw)
            if plot.src.endswith('.pdf'):
                page.img(class_='img-responsive',
                         src=plot.src.replace('.pdf', '.png'))
            else:
                page.img(class_='img-responsive', src=plot.src)
            page.a.close()
            page.div.close()
            # detect end of row
            if (i + 1) == layout[l][0]:
                i = 0
                l += 1
                page.div.close()
            # detect last plot
            elif j == (len(plots) - 1):
                page.div.close()
                break
            # or move to next column
            else:
                i += 1

        page.div.close()
        return page

    def html_content(self, content):
        page = markup.page()
        if self.foreword:
            page.add(str(self.foreword))
        if content:
            page.add(str(content))
        page.add(str(self.scaffold_plots()))
        if self.afterword:
            page.add(str(self.afterword))
        return Tab.html_content(str(page))

    def write_html(self, foreword=None, afterword=None, **kwargs):
        """Write the HTML page for this tab.

        Parameters
        ----------
        foreword : `str`, :class:`~MarkupPy.markup.page`, optional
            content to place above the plot grid, defaults to
            :attr:`PlotTab.foreword`
        afterword : `str`, :class:`~MarkupPy.markup.page`, optional
            content to place below the plot grid, defaults to
            :attr:`PlotTab.afterword`
        **kwargs
            other keyword arguments to be passed through
            :meth:`~Tab.write_html`

        See Also
        --------
        gwsumm.tabs.Tab.write_html : for details of all valid unnamed
                                     keyword arguments
        """
        if not kwargs.pop('writehtml', True):
            return
        if foreword is not None:
            self.foreword = foreword
        if afterword is not None:
            self.afterword = self.afterword
        return super(PlotTab, self).write_html(None, **kwargs)


register_tab(PlotTab)


class StateTab(PlotTab):
    """Tab with multiple content pages defined via 'states'

    Each state is printed to its own HTML file which is loaded via
    javascript upon request into the #main div of the index for the tab.

    Parameters
    ----------
    name : `str`
        name of this tab (required)
    states : `list`
        a list of states for this tab. Each state can take any form,
        but must be castable to a `str` in order to be printed.
    plots : `list`
        list of plots to display on this tab. More plots can be added
        at any time via :meth:`PlotTab.add_plot`
    layout : `int`, `list`
        the number of plots to display in each row, or a list of numbers
        to define each row individually. If the number of plots defined
        by the layout is less than the total number of plots, the layout
        for the final row will be repeated as necessary.

        For example ``layout=[1, 2, 3]`` will display a single plot on
        the top row, two plots on the second, and 3 plots on each row
        thereafter.
    index : `str`
        HTML file in which to write. By default each tab is written to
        an index.html file in its own directory. Use :attr:`~Tab.index`
        to find out the default index, if not given.
    shortname : `str`
        shorter name for this tab to use in the navigation bar. By
        default the regular name is used
    parent : :class:`~gwsumm.tabs.Tab`
        parent of this tab. This is used to position this tab in the
        navigation bar.
    children : `list`
        list of child :class:`Tabs <~gwsumm.tabs.Tab>` of this one. This
        is used to position this tab in the navigation bar.
    group : `str`
        name of containing group for this tab in the navigation bar
        dropdown menu. This is only relevant if this tab has a parent.
    path : `str`
        base output directory for this tab (should be the same directory
        for all tabs in this run).
    """
    type = 'state'

    def __init__(self, name, states=list(), **kwargs):
        """Initialise a new `Tab`.
        """
        if kwargs.get('mode', None) is None:
            raise ValueError("%s() needs keyword argument 'mode'"
                             % type(self).__name__)
        super(StateTab, self).__init__(name, **kwargs)
        # process states
        if not isinstance(states, (tuple, list)):
            states = [states]
        self.states = states

    # -------------------------------------------
    # StateTab properties

    @property
    def states(self):
        """The set of :class:`states <gwsumm.state.SummaryState>` over
        whos times this tab's data will be processed.

        The set of states will be linked in the given order with a switch
        on the far-right of the HTML navigation bar.
        """
        return self._states

    @states.setter
    def states(self, statelist, default=None):
        self._states = []
        for state in statelist:
            # allow default indication by trailing asterisk
            if state == default:
                default_ = True
            elif (default is None and isinstance(state, string_types) and
                    state.endswith('*')):
                state = state[:-1]
                default_ = True
            else:
                default_ = False
            self.add_state(state, default=default_)

    def add_state(self, state, default=False):
        """Add a `SummaryState` to this tab

        Parameters
        ----------
        state : `str`, :class:`~gwsumm.state.SummaryState`
            either the name of a state, or a `SummaryState`.
        register : `bool`, default: `False`
            automatically register all new states
        """
        if not isinstance(state, SummaryState):
            state = get_state(state)
        self._states.append(state)
        if default:
            self.defaultstate = state

    @property
    def defaultstate(self):
        try:
            return self._defaultstate
        except AttributeError:
            self._defaultstate = self.states[0]
            return self._defaultstate

    @defaultstate.setter
    def defaultstate(self, state):
        self._defaultstate = state

    @property
    def frames(self):
        # write page for each state
        statelinks = []
        outdir = os.path.split(self.index)[0]
        for i, state in enumerate(self.states):
            statelinks.append(os.path.join(
                outdir, '%s.html' % re_cchar.sub('_', str(state).lower())))
        return statelinks

    # -------------------------------------------
    # StateTab methods

    @classmethod
    def from_ini(cls, cp, section, *args, **kwargs):
        # parse states and retrieve their definitions
        if cp.has_option(section, 'states'):
            # states listed individually
            kwargs.setdefault(
                'states', [re_quote.sub('', s).strip() for s in
                           cp.get(section, 'states').split(',')])
        else:
            # otherwise use 'all' state - full span with no gaps
            kwargs.setdefault('states', ['All'])
        # parse core Tab information
        return super(StateTab, cls).from_ini(cp, section, *args, **kwargs)

    # ------------------------------------------------------------------------
    # HTML methods

    def html_navbar(self, brand='', **kwargs):
        """Build the navigation bar for this `Tab`.

        The navigation bar will consist of a switch for this page linked
        to other interferometer servers, followed by the navbar brand,
        then the full dropdown-based navigation menus configured for the
        given ``tabs`` and their descendents.

        Parameters
        ----------
        brand : `str`, :class:`~MarkupPy.markup.page`
            content for navbar-brand
        ifo : `str`, optional
            prefix for this IFO.
        ifomap : `dict`, optional
            `dict` of (ifo, {base url}) pairs to map to summary pages for
            other IFOs.

        tabs : `list`, optional
            list of parent tabs (each with a list of children) to include
            in the navigation bar.

        Returns
        -------
        page : `~MarkupPy.markup.page`
            a markup page containing the navigation bar.
        """
        brand_ = markup.page()
        # build state switch
        if len(self.states) > 1 or str(self.states[0]) != ALLSTATE:
            default = self.states.index(self.defaultstate)
            brand_.add(str(html.state_switcher(
                list(zip(self.states, self.frames)), default)))
        # build HTML brand
        if brand:
            brand_.add(str(brand))
        # combine and return
        return super(StateTab, self).html_navbar(brand=brand_, **kwargs)

    @staticmethod
    def html_content(frame):
        r"""Build the #main div for this tab.

        In this construction, the <div id="id\_"> is empty, with a
        javascript hook to load the given frame into the div when ready.
        """
        wrappedcontent = html.load(frame, id_='content')
        return Tab.html_content(str(wrappedcontent))

    def write_state_html(self, state, pre=None, post=None, plots=True):
        """Write the frame HTML for the specific state of this tab

        Parameters
        ----------
        state : `~gwsumm.state.SummaryState`
            `SummaryState` over which to generate inner HTML
        """
        # build page
        page = markup.page()
        if pre:
            page.add(str(pre))
        if plots:
            page.add(str(self.scaffold_plots(state=state)))
        if post:
            page.add(str(post))
        # write to file
        idx = self.states.index(state)
        with open(self.frames[idx], 'w') as fobj:
            fobj.write(str(page))
        return self.frames[idx]

    def write_html(self, title=None, subtitle=None, tabs=list(), ifo=None,
                   ifomap=dict(), brand=None, css=html.CSS, js=html.JS,
                   about=None, footer=None, **inargs):
        """Write the HTML page for this state Tab.

        Parameters
        ----------
        maincontent : `str`, :class:`~MarkupPy.markup.page`
            simple string content, or a structured `page` of markup to
            embed as the content of the #main div.
        title : `str`, optional, default: {parent.name}
            level 1 heading for this `Tab`.
        subtitle : `str`, optional, default: {self.name}
            level 2 heading for this `Tab`.
        tabs: `list`, optional
            list of top-level tabs (with children) to populate navbar
        ifo : `str`, optional
            prefix for this IFO.
        ifomap : `dict`, optional
            `dict` of (ifo, {base url}) pairs to map to summary pages for
            other IFOs.
        brand : `str`, :class:`~MarkupPy.markup.page`, optional
            non-menu content for navigation bar, defaults to calendar
        css : `list`, optional
            list of resolvable URLs for CSS files. See `gwsumm.html.CSS` for
            the default list.
        js : `list`, optional
            list of resolvable URLs for javascript files. See
            `gwumm.html.JS` for the default list.
        about : `str`, optional
            href for the 'About' page
        footer : `str`, `~MarkupPy.markup.page`
            user-defined content for the footer (placed below everything else)
        **inargs
            other keyword arguments to pass to the
            :meth:`~Tab.build_inner_html` method
        """
        default = self.states.index(self.defaultstate)
        return super(PlotTab, self).write_html(
            self.frames[default], title=title, subtitle=subtitle,
            tabs=tabs, ifo=ifo, ifomap=ifomap, brand=brand, css=css, js=js,
            about=about, footer=footer, **inargs)


register_tab(StateTab)


class UrlTab(Tab):
    type = 'link'

    def __init__(self, name, url, **kwargs):
        super(UrlTab, self).__init__(name, **kwargs)
        self.href = url

    @property
    def href(self):
        return self._href

    @href.setter
    def href(self, url):
        self._href = url

    @classmethod
    def from_ini(cls, cp, section, *args, **kwargs):
        kwargs.setdefault('url', cp.get(section, 'url'))
        return super(UrlTab, cls).from_ini(cp, section, *args, **kwargs)

    def write_html(self, **kwargs):
        return


register_tab(UrlTab)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Registry for GWSumm output plot types

All plot types should be registered for easy identification from the
configuration INI files
"""

import re
from six import string_types

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

_PLOTS = {}


def register_plot(plot, name=None, force=False):
    """Register a new summary `Plot` to the given ``name``

    Parameters
    ----------
    name : `str`
        unique descriptive name for this type of plot, must not
        contain any spaces, e.g. 'timeseries'
    plotclass : `type`
        defining Class for this plot type
    force : `bool`
        overwrite existing registration for this type

    Raises
    ------
    ValueError
        if name is already registered and ``force`` not given as `True`
    """
    if name is None:
        name = plot.type
    if name not in _PLOTS or force:
        _PLOTS[name] = plot
    else:
        raise ValueError("Plot '%s' has already been registered to the %s "
                         "class" % (name, plot.__name__))


def get_plot(name):
    """Query the registry for the plot class registered to the given
    name
    """
    if isinstance(name, string_types):
        name = re.sub(r'[\'\"]', '', name)
    try:
        return _PLOTS[name]
    except KeyError:
        raise ValueError("No TabPlot registered with name '%s'" % name)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Get spectrograms and spectra
"""

from __future__ import division

import os.path
import operator
import warnings
from collections import OrderedDict

from six import string_types
from six.moves import reduce

# imports for filter
from math import pi  # noqa: F401

import numpy

from scipy import interpolate

from astropy import units

from gwpy.segments import DataQualityFlag
from gwpy.frequencyseries import FrequencySeries
from gwpy.spectrogram import SpectrogramList

from .. import (globalv, io)
from ..utils import (vprint, safe_eval)
from ..channels import (
    get_channel,
    split_combination as split_channel_combination,
)
from .utils import (use_segmentlist, make_globalv_key, get_fftparams)
from .mathutils import (get_with_math, parse_math_definition)
from .timeseries import (get_timeseries, get_timeseries_dict)

OPERATOR = {
    '*': operator.mul,
    '-': operator.sub,
    '+': operator.add,
    '/': operator.truediv,
}

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- spectrogram --------------------------------------------------------------

@use_segmentlist
def get_spectrogram(channel, segments, config=None, cache=None,
                    query=True, nds=None, format='power', return_=True,
                    frametype=None, nproc=1, datafind_error='raise',
                    **fftparams):
    """Retrieve the time-series and generate a spectrogram of the given
    channel
    """
    channel = get_channel(channel)

    # read data for all sub-channels
    specs = []
    channels = split_channel_combination(channel)
    for c in channels:
        specs.append(_get_spectrogram(c, segments, config=config, cache=cache,
                                      query=query, nds=nds, format=format,
                                      return_=return_, frametype=frametype,
                                      nproc=nproc,
                                      datafind_error=datafind_error,
                                      **fftparams))
    if return_ and len(channels) == 1:
        return specs[0]
    elif return_:
        return get_with_math(
            channel, segments, _get_spectrogram, _get_spectrogram,
            config=config, query=False, format=format, return_=True,
            **fftparams)


@use_segmentlist
def _get_spectrogram(channel, segments, config=None, cache=None,
                     query=True, nds=None, format='power', return_=True,
                     frametype=None, nproc=1,
                     datafind_error='raise', **fftparams):
    channel = get_channel(channel)

    # if we aren't given a method, check to see whether data have already
    # been processed, if so, choose that one
    if fftparams.get('method', None) is None:
        methods = set([key.split(';')[1] for key in globalv.SPECTROGRAMS
                       if key.startswith('%s;' % channel.ndsname)])
        try:
            fftparams['method'] = list(methods)[0]
        except IndexError:
            fftparams['method'] = 'welch'

    # clean fftparams dict using channel default values
    fftparams = get_fftparams(channel, **fftparams)
    # override special-case methods
    if format in ['rayleigh']:
        fftparams.method = format

    # key used to store the coherence spectrogram in globalv
    key = make_globalv_key(channel, fftparams)

    # keep FftParams as a dict for convenience
    fftparams = fftparams.dict()

    # read segments from global memory
    havesegs = globalv.SPECTROGRAMS.get(key, SpectrogramList()).segments
    new = segments - havesegs
    query &= abs(new) != 0

    globalv.SPECTROGRAMS.setdefault(key, SpectrogramList())

    if query:
        # extract spectrogram stride from dict
        try:
            stride = float(fftparams.pop('stride'))
        except (TypeError, KeyError) as e:
            msg = ('cannot parse a spectrogram stride from the kwargs '
                   'given, please give some or all of fftlength, overlap, '
                   'stride')
            if isinstance(e, TypeError):
                e.args = (msg,)
                raise
            raise TypeError(msg)

        # read channel information
        try:
            filter_ = channel.frequency_response
        except AttributeError:
            filter_ = None
        else:
            if isinstance(filter_, string_types) and os.path.isfile(filter_):
                filter_ = io.read_frequencyseries(filter_)
            elif isinstance(filter_, string_types):
                filter_ = safe_eval(filter_, strict=True)

        # get time-series data
        timeserieslist = get_timeseries(channel, new, config=config,
                                        cache=cache, frametype=frametype,
                                        nproc=nproc, query=query,
                                        datafind_error=datafind_error, nds=nds)
        # calculate spectrograms
        if len(timeserieslist):
            vprint("    Calculating (%s) spectrograms for %s"
                   % (fftparams['method'], str(channel)))
        for ts in timeserieslist:
            # if too short for a single segment, continue
            if abs(ts.span) < (stride + fftparams.get('overlap', 0)):
                continue
            # truncate timeseries to integer number of strides
            d = size_for_spectrogram(ts.duration.to('s').value, stride,
                                     fftparams['fftlength'],
                                     fftparams.get('overlap', 0))
            ts = ts.crop(ts.span[0], ts.span[0] + d, copy=False)
            # calculate spectrogram
            try:
                # rayleigh spectrogram has its own instance method
                if fftparams.get('method', None) == 'rayleigh':
                    spec_kw = fftparams.copy()
                    for fftkey in ('method', 'scheme',):  # remove ASD keys
                        spec_kw.pop(fftkey, None)
                    spec_func = ts.rayleigh_spectrogram
                else:
                    spec_kw = fftparams
                    spec_func = ts.spectrogram
                specgram = spec_func(stride, nproc=nproc, **spec_kw)
            except ZeroDivisionError:
                if stride == 0:
                    raise ZeroDivisionError("Spectrogram stride is 0")
                elif fftparams['fftlength'] == 0:
                    raise ZeroDivisionError("FFT length is 0")
                else:
                    raise
            except ValueError as e:
                if 'has no unit' in str(e):
                    unit = ts.unit
                    ts._unit = units.Unit('count')
                    specgram = ts.spectrogram(stride, nproc=nproc, **fftparams)
                    specgram._unit = unit ** 2 / units.Hertz
                else:
                    raise
            if isinstance(filter_, FrequencySeries) and (
                    fftparams['method'] not in ['rayleigh']):
                specgram = apply_transfer_function_series(specgram, filter_)
            elif filter_ and fftparams['method'] not in ['rayleigh']:
                # manually setting x0 is a hack against precision error
                # somewhere inside the **(1/2.) operation (Quantity)
                x0 = specgram.x0.value
                specgram = (specgram ** (1/2.)).filter(*filter_,
                                                       inplace=True) ** 2
                specgram.x0 = x0
            if specgram.unit is None:
                specgram._unit = channel.unit
            elif len(globalv.SPECTROGRAMS[key]):
                specgram._unit = globalv.SPECTROGRAMS[key][-1].unit
            add_spectrogram(specgram, key=key)
            vprint('.')
        if len(timeserieslist):
            vprint('\n')

    if not return_:
        return

    # return correct data
    out = SpectrogramList()
    for specgram in globalv.SPECTROGRAMS[key]:
        for seg in segments:
            if abs(seg) < specgram.dt.value:
                continue
            if specgram.span.intersects(seg):
                common = specgram.span & type(seg)(seg[0],
                                                   seg[1] + specgram.dt.value)
                s = specgram.crop(*common)
                if format in ['amplitude', 'asd']:
                    s = s**(1/2.)
                elif format in ['rayleigh']:
                    # XXX FIXME: this corrects the bias offset in Rayleigh
                    med = numpy.median(s.value)
                    s /= med
                if s.shape[0]:
                    out.append(s)
    return out.coalesce()


def add_spectrogram(specgram, key=None, coalesce=True):
    """Add a `Spectrogram` to the global memory cache
    """
    if key is None:
        key = specgram.name or str(specgram.channel)
    globalv.SPECTROGRAMS.setdefault(key, SpectrogramList())
    globalv.SPECTROGRAMS[key].append(specgram)
    if coalesce:
        globalv.SPECTROGRAMS[key].coalesce()


@use_segmentlist
def get_spectrograms(channels, segments, config=None, cache=None, query=True,
                     nds=None, format='power', return_=True, frametype=None,
                     nproc=1, datafind_error='raise', **fftparams):
    """Get spectrograms for multiple channels
    """
    channels = list(map(get_channel, channels))

    # get timeseries data in bulk
    if query:
        # get underlying list of data channels to read
        qchannels = list(map(get_channel,
                        set([c for group in
                             map(split_channel_combination, channels)
                             for c in group])))

        # work out FFT params and storage keys for each data channel
        keys = []
        for channel in qchannels:
            fftparams_ = get_fftparams(channel, **fftparams)
            keys.append(make_globalv_key(channel, fftparams_))

        # restrict segments to those big enough to hold >= 1 stride
        strides = set([getattr(c, 'stride', 0) for c in qchannels])
        if len(strides) == 1:
            stride = strides.pop()
            segments = type(segments)(s for s in segments if abs(s) >= stride)

        # work out new segments for which to read data
        havesegs = reduce(operator.and_, (globalv.SPECTROGRAMS.get(
            key, SpectrogramList()).segments for key in keys))
        new = segments - havesegs

        # read data for new segments
        get_timeseries_dict(qchannels, new, config=config, cache=cache,
                            nproc=nproc, frametype=frametype,
                            datafind_error=datafind_error, nds=nds,
                            return_=False)
    # loop over channels and generate spectrograms
    out = OrderedDict()
    for channel in channels:
        out[channel] = get_spectrogram(
            channel, segments, config=config, cache=cache, query=query,
            nds=nds, format=format, nproc=nproc,
            return_=return_, datafind_error=datafind_error,
            **fftparams)
    return out


def size_for_spectrogram(size, stride, fftlength, overlap):
    if size < stride:
        return None
    x = size // stride * stride + overlap
    if x > size:
        x -= fftlength
    return x


def apply_transfer_function_series(specgram, tfunc):
    """Multiply a spectrogram by a transfer function `FrequencySeries`

    This method interpolates the transfer function onto the frequency vector
    of the spectrogram, so should work regardless of the inputs
    """
    # interpolate transfer function onto spectrogram frequency series
    interpolator = interpolate.interp1d(tfunc.frequencies.value, tfunc.value)
    itfunc = numpy.zeros((1, specgram.frequencies.size))
    known = specgram.frequencies.value >= tfunc.frequencies.value[0]
    known &= specgram.frequencies.value <= tfunc.frequencies.value[-1]
    itfunc[0, :][known] = interpolator(specgram.frequencies.value[known])
    # and multiply
    return (specgram ** (1/2.) * itfunc) ** 2


# -- spectrum -----------------------------------------------------------------

@use_segmentlist
def get_spectrum(channel, segments, config=None, cache=None,
                 query=True, nds=None, format='power', return_=True,
                 frametype=None, nproc=1, datafind_error='raise',
                 **fftparams):
    """Retrieve the time-series and generate a spectrogram of the given
    channel
    """
    channel = get_channel(channel)
    if isinstance(segments, DataQualityFlag):
        name = ','.join([channel.ndsname, segments.name])
        segments = segments.active
    else:
        name = channel.ndsname
    name += ',%s' % format

    # read data for all sub-channels
    specs = []
    channels = list(parse_math_definition(str(channel))[0])
    if len(channels) == 0:
        channels = [channel]
    for c in channels:
        specs.append(_get_spectrum(c, segments, config=config, cache=cache,
                                   query=query, nds=nds, format=format,
                                   return_=return_, frametype=frametype,
                                   nproc=nproc,
                                   datafind_error=datafind_error,
                                   **fftparams))
    if return_ and len(channels) == 1:
        return specs[0]
    elif return_:
        return [get_with_math(
                    channel, segments, _get_spectrum, _get_spectrum,
                    config=config, format=format, return_=True,
                    which=which)[0] for which in ['mean', 'min', 'max']]


def _get_spectrum(channel, segments, config=None, cache=None, query=True,
                  nds=None, format='power', return_=True, which='all',
                  **fftparams):
    """Retrieve the time-series and generate a spectrogram of the given
    channel
    """
    channel = get_channel(channel)
    if isinstance(segments, DataQualityFlag):
        name = ','.join([channel.ndsname, segments.name])
        segments = segments.active
    else:
        name = channel.ndsname
    name += ',%s' % format
    cmin = '%s.min' % name
    cmax = '%s.max' % name

    if name not in globalv.SPECTRUM:
        if os.path.isfile(channel.ndsname):
            globalv.SPECTRUM[name] = io.read_frequencyseries(channel.ndsname)
            globalv.SPECTRUM[cmin] = globalv.SPECTRUM[name]
            globalv.SPECTRUM[cmax] = globalv.SPECTRUM[name]
        else:
            fftparams.setdefault('fftlength', 1)
            fftparams.setdefault('overlap', 0.5)
            if 'stride' not in fftparams and 'fftlength' in fftparams:
                fftparams.setdefault('stride', fftparams['fftlength'])

            speclist = get_spectrogram(channel, segments, config=config,
                                       cache=cache, query=query, nds=nds,
                                       format=format, **fftparams)
            try:
                specgram = speclist.join(gap='ignore')
            except ValueError as e:
                if 'units do not match' in str(e):
                    warnings.warn(str(e))
                    for spec in speclist[1:]:
                        spec.unit = speclist[0].unit
                    specgram = speclist.join(gap='ignore')
                else:
                    raise
            try:
                globalv.SPECTRUM[name] = specgram.percentile(50)
            except (ValueError, IndexError):
                globalv.SPECTRUM[name] = FrequencySeries(
                    [], channel=channel, f0=0, df=1, unit=units.Unit(''))
                globalv.SPECTRUM[cmin] = globalv.SPECTRUM[name]
                globalv.SPECTRUM[cmax] = globalv.SPECTRUM[name]
            else:
                globalv.SPECTRUM[cmin] = specgram.percentile(5)
                globalv.SPECTRUM[cmax] = specgram.percentile(95)

    if not return_:
        return

    if which == 'all':
        return (globalv.SPECTRUM[name], globalv.SPECTRUM[cmin],
                globalv.SPECTRUM[cmax])
    if which == 'mean':
        return globalv.SPECTRUM[name]
    if which == 'min':
        return globalv.SPECTRUM[cmin]
    if which == 'max':
        return globalv.SPECTRUM[cmax]
    raise ValueError("Unrecognised value for `which`: %r" % which)



# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""Custom `SummaryTab` for the output of the FScan algorithm.
"""

import os
import glob

from six import string_types

from dateutil import parser

from MarkupPy import markup

from .registry import (get_tab, register_tab)

from .. import html
from .. plot import get_plot
from ..mode import Mode
from ..config import GWSummConfigParser

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['FscanTab']

base = get_tab('default')
SummaryPlot = get_plot(None)


class FscanTab(base):
    """Custom tab displaying a summary of Fscan results.
    """
    type = 'fscan'

    def __init__(self, *args, **kwargs):
        if kwargs['mode'] != Mode.day:
            raise RuntimeError("FscanTab is only available in %s mode."
                               % Mode.day.name)
        super(FscanTab, self).__init__(*args, **kwargs)

    @classmethod
    def from_ini(cls, config, section, **kwargs):
        """Define a new `FscanTab` from a `ConfigParser`.
        """
        # parse generic configuration
        new = super(FscanTab, cls).from_ini(config, section, **kwargs)
        new.set_layout([2])

        # work out day directory and url
        new.channel = config.get(section, 'channel')
        base = os.path.normpath(config.get(section, 'base-directory'))
        dirs = glob.glob(base)
        if len(dirs) == 0:
            d = None
        elif len(dirs) > 1:
            d = dirs
        else:
            d = dirs[0]
        if d is not None:
            new.directory = os.path.join(d, new.channel.replace(':', '_'))
        else:
            new.directory = None

        # get navigation urls
        new.navigation = []
        for key, val in filter(lambda x: x[0].startswith('navigation'),
                               config.items(section)):
            system = key.split('-', 1)[1]
            new.navigation.append((system, val))

        return new

    def process(self, config=GWSummConfigParser(), **kwargs):
        # find all plots
        self.plots = []
        if isinstance(self.directory, string_types):
            plots = sorted(
                glob.glob(os.path.join(self.directory, 'spec_*.png')),
                key=lambda p: float(os.path.basename(p).split('_')[1]))
            for p in plots:
                home_, postbase = p.split('/public_html/', 1)
                user = os.path.split(home_)[1]
                self.plots.append(SummaryPlot(
                    href='/~%s/%s' % (user, postbase)))

    def write_state_html(self, state):
        """Write the '#main' HTML content for this `FscanTab`.
        """
        page = markup.page()

        if self.directory is None:
            page.div(class_='alert alert-warning')
            page.p("No analysis was performed for this period, "
                   "please try again later.")
            page.p("If you believe these data should have been found, please "
                   "contact %s."
                   % markup.oneliner.a('the DetChar group',
                                       class_='alert-link',
                                       href='mailto:detchar@ligo.org'))
            page.div.close()
            self.directory = []
        elif isinstance(self.directory, list):
            page.div(class_='alert alert-warning')
            page.p("Multiple results directories were found for this period, "
                   "cannot determine correct directory.")
            page.p("If you believe this to be an error, please contact %s."
                   % markup.oneliner.a('the DetChar group',
                                       class_='alert-link',
                                       href='mailto:detchar@ligo.org'))
            page.div.close()
            page.hr(class_='row-divider')
        elif not self.plots:
            page.div(class_='alert alert-warning')
            page.p("This analysis produced no plots.")
            page.p("If you believe these data should have been found, please "
                   "contact %s."
                   % markup.oneliner.a('the DetChar group',
                                       class_='alert-link',
                                       href='mailto:detchar@ligo.org'))
            page.div.close()
            page.hr(class_='row-divider')
            self.directory = [self.directory]
        else:
            self.directory = [self.directory]

        # link full results
        for d in self.directory:
            # parse date
            datestr = d.split(os.path.sep)[-2].split('_')[1:]
            datestr = '{0}-{1}-{2} {3}:{4}:{5} {6}'.format(*datestr)
            date = parser.parse(datestr).strftime('%Y-%m-%d %H:%M:%S %Z')
            # find HTML
            try:
                index = glob.glob(os.path.join(
                            d, self.channel.replace(':', '_', 1), '*html'))[0]
            except IndexError:
                index = d + os.path.sep
            home_, postbase = index.split('/public_html/', 1)
            user = os.path.split(home_)[1]
            index = '/~%s/%s' % (user, postbase)
            page.div(class_='btn-group')
            page.a('Click here for the full Fscan results for %s' % date,
                   href=index, rel='external', target='_blank',
                   class_='btn btn-default btn-info btn-xl')
            page.div.close()

        if self.navigation:
            page.h2('Fscan links:')
        for i, (key, url) in enumerate(sorted(self.navigation,
                                              key=lambda x: x[0])):
            if not i % 6:
                page.div(class_="row")
            if i + 6 > len(self.navigation):
                page.div(class_='col-sm-2')
            else:
                page.div(class_='col-sm-2', style='margin-bottom: 10px;')
            page.a(key, href=url, rel='external', target='_blank',
                   class_='btn btn-success btn-lg btn-block')
            page.div.close()
            if i % 6 == 6 or i == len(self.navigation) - 1:
                page.div.close()

        if self.plots:
            page.hr(class_='row-divider')
            page.div(class_='scaffold well')
            # parse frequencies
            freqs = [list(map(float, os.path.basename(p.href).split('_')[1:3]))
                     for p in self.plots[::2]]
            # build a grid of buttons as frequency markers
            ncols = 5
            linksize = 12 // ncols
            offset = 1
            i = 0
            page.p("Select a frequency range to jump to those plots:")
            for i, (f1, f2) in enumerate(freqs):
                if i % ncols == 0:
                    page.div(class_='row', style="margin-bottom: 10px;")
                if offset and i % ncols == 0:
                    page.div(class_="col-md-%d col-md-offset-%d"
                                    % (linksize, offset))
                else:
                    page.div(class_="col-md-%d" % linksize)
                page.div(class_="btn-group btn-group-justified")
                page.a('%s-%s Hz' % (f1, f2), href="javascript:;",
                       onclick="document.location.hash='%s-%s'" % (f1, f2),
                       role="button", type="button", class_="btn btn-warning")
                page.div.close()
                page.div.close()
                if i % ncols == ncols - 1:
                    page.div.close()
            if i % ncols != ncols - 1:
                page.div.close()
            page.hr(class_='row-divider')
            # reverse frequency order
            spectrograms = self.plots[::2]
            spectra = self.plots[1::2]
            for pair in list(zip(spectrograms, spectra, freqs))[::-1]:
                f = pair[-1]
                page.div(class_='row', id="%s-%s" % (f[0], f[1]))
                for p in pair[:2]:
                    page.div(class_="col-md-6")
                    page.a(href=p.href, class_="fancybox plot",
                           **{'data-fancybox-group': 1})
                    page.img(class_='img-responsive', src=p.href)
                    page.a.close()
                    page.div.close()
                page.div.close()
            page.div.close()

        # write to file
        idx = self.states.index(state)
        with open(self.frames[idx], 'w') as fobj:
            fobj.write(str(page))
        return self.frames[idx]


register_tab(FscanTab)
